"""
..
   This file is part of the CoCy program.
   Copyright (C) 2011 Michael N. Lipp
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

.. codeauthor:: mnl
"""
from circuits.web.controllers import BaseController, expose
import os
from circuits_bricks.web import ScopedChannel
from xml.etree import ElementTree
from cocy.upnp import UPNP_SERVICE_SCHEMA
from cocy.misc import set_ns_prefixes


class UPnPService(BaseController):
    """
    A ``UPnPService`` component holds information about a service
    and provides HTTP access to the UPnP service description.
    
    As specified in the UPnP architecture, services are
    identified by a type and a version. The service descriptions are made
    available to the clients via HTTP GET requests to specific URLs 
    that return the description as XML text.    
    The URL for accessing a device's service description is part of the
    information provided for the device.
    """

    channel = None
    _template_dir = os.path.join(os.path.dirname(__file__), "templates")
    _service_dir = os.path.join(os.path.dirname(__file__), "services")

    def __init__(self, config_id, type_ver):
        """
        Instances of this component are created from a service
        description that is selected by the given type and version.        
        The service description is looked up in a directory
        where it is stored as a file named ``type_ver.xml``.
        
        :param type: the service type
        :param ver: the service version
        """
        self._type, self._ver = type_ver.split(":", 1)
        self._path = "/%s_%s" % (self._type, self._ver)
        self.channel = ScopedChannel("upnp-web", self._path)
        # Now call super as only now the channel is known and this classes
        # handlers will be registered properly
        super(UPnPService, self).__init__();
        sfile = open(os.path.join(self._service_dir, 
                                  "%s_%s.xml" % (self._type, self._ver)))
        sd = ElementTree.parse(sfile).getroot()
        sd.set("configId", str(config_id))
        # Some Android clients have problems with white spaces
        for el in sd.getiterator():
            if el.text:
                el.text = el.text.strip()
            if el.tail:
                el.tail = el.tail.strip()
        set_ns_prefixes(sd, { "": UPNP_SERVICE_SCHEMA })
        self._description = ElementTree.tostring(sd)

    @property
    def type(self):
        return self._type

    @property
    def type_ver(self):
        return "%s:%s" % (self._type, str(self._ver))

    @property
    def description_url(self):
        return self._path + "/service.xml" 

    @expose("service.xml")
    def _on_description(self, *args):
        self.response.headers["Content-Type"] = "text/xml"
        return self._description



#!/usr/bin/env python
#
# soaplib - Copyright (C) cocy.soaplib contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

import datetime
import unittest

from lxml import etree

from cocy.soaplib.serializers.clazz import ClassSerializer
from cocy.soaplib.serializers.clazz import Array
from cocy.soaplib.serializers.primitive import DateTime
from cocy.soaplib.serializers.primitive import Float
from cocy.soaplib.serializers.primitive import Integer
from cocy.soaplib.serializers.primitive import String

from cocy.soaplib import service
from cocy.soaplib import wsgi
from cocy.soaplib.service import rpc

class Address(ClassSerializer):
    __namespace__ = "TestService"

    street = String
    city = String
    zip = Integer
    since = DateTime
    laditude = Float
    longitude = Float

class Person(ClassSerializer):
    __namespace__ = "TestService"

    name = String
    birthdate = DateTime
    age = Integer
    addresses = Array(Address)
    titles = Array(String)

class Request(ClassSerializer):
    __namespace__ = "TestService"

    param1 = String
    param2 = Integer

class Response(ClassSerializer):
    __namespace__ = "TestService"

    param1 = Float

class TypeNS1(ClassSerializer):
    __namespace__ = "TestService.NS1"

    s = String
    i = Integer

class TypeNS2(ClassSerializer):
    __namespace__ = "TestService.NS2"

    d = DateTime
    f = Float

class MultipleNamespaceService(service.DefinitionBase):
    @rpc(TypeNS1, TypeNS2)
    def a(self, t1, t2):
        return "OK"

class MultipleNamespaceValidatingService(MultipleNamespaceService):
    def __init__(self):
        MultipleNamespaceService.__init__(self)

        self.validating_service = True

class TestService(service.DefinitionBase):
    @rpc(String, _returns=String)
    def aa(self, s):
        return s

    @rpc(String, Integer, _returns=DateTime)
    def a(self, s, i):
        return datetime.datetime.now()

    @rpc(Person, String, Address, _returns=Address)
    def b(self, p, s, a):
        return Address()

    @rpc(Person, isAsync=True)
    def d(self, Person):
        pass

    @rpc(Person, isCallback=True)
    def e(self, Person):
        pass

    @rpc(String, String, String, _returns=String,
        _in_variable_names={'_from': 'from', '_self': 'self',
            '_import': 'import'},
        _out_variable_name="return")
    def f(self, _from, _self, _import):
        return '1234'

class MultipleReturnService(service.DefinitionBase):
    @rpc(String, _returns=(String, String, String))
    def multi(self, s):
        return s, 'a', 'b'

class Test(unittest.TestCase):
    '''
    Most of the service tests are excersized through the interop tests
    '''

    def setUp(self):
        self.app = wsgi.Application([TestService], 'tns')
        self.srv = TestService()
        self._wsdl = self.app.get_wsdl('')
        self.wsdl = etree.fromstring(self._wsdl)

    def test_portypes(self):
        porttype = self.wsdl.find('{http://schemas.xmlsoap.org/wsdl/}portType')
        self.assertEquals(
            len(self.srv.public_methods), len(porttype.getchildren()))

    def test_override_param_names(self):
        for n in ['self', 'import', 'return', 'from']:
            self.assertTrue(n in self._wsdl, '"%s" not in self._wsdl' % n)

    def test_multiple_return(self):
        app = wsgi.Application([MultipleReturnService], 'tns')
        app.get_wsdl('')
        srv = MultipleReturnService()
        message = srv.public_methods[0].out_message()

        self.assertEquals(len(message._type_info), 3)

        sent_xml = etree.Element('test')
        message.to_xml( ('a','b','c'), srv.get_tns(), sent_xml )
        sent_xml = sent_xml[0]

        print etree.tostring(sent_xml, pretty_print=True)
        response_data = message.from_xml(sent_xml)

        self.assertEquals(len(response_data), 3)
        self.assertEqual(response_data[0], 'a')
        self.assertEqual(response_data[1], 'b')
        self.assertEqual(response_data[2], 'c')

    def test_multiple_ns(self):
        svc = wsgi.Application([MultipleNamespaceService], 'tns')
        wsdl = svc.get_wsdl("URL")

if __name__ == '__main__':
    unittest.main()



"""
..
   This file is part of the CoCy program.
   Copyright (C) 2011 Michael N. Lipp
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from cocy.upnp.adapters.adapter import upnp_service, UPnPServiceController,\
    upnp_state, Notification, UPnPServiceError
from circuits_bricks.app.logger import Log
import logging
from time import time
from circuits_bricks.core.timers import Timer
from circuits.core.events import Event
from circuits.core.handlers import handler
from StringIO import StringIO
from xml.etree.ElementTree import Element, QName, ElementTree, SubElement
from cocy.upnp import UPNP_AVT_EVENT_NS, UPNP_RCS_EVENT_NS
from cocy import misc

class UPnPCombinedEventsServiceController(UPnPServiceController):
    
    def __init__(self, adapter, device_path, service, service_id, event_ns):
        super(UPnPCombinedEventsServiceController, self).__init__\
            (adapter, device_path, service, service_id)
        self._event_ns = event_ns
        self._changes = dict()
        self._updates_locked = False
        
    @upnp_state(evented_by=None)
    def LastChange(self):
        root = Element(QName(self._event_ns, "Event"))
        inst = SubElement(root, QName(self._event_ns, "InstanceID"), 
                                      { "val": "0" })
        for name, value in self._changes.items():
            SubElement(inst, QName(self._event_ns, name), { "val": value })
        misc.set_ns_prefixes(root, { "": self._event_ns })
        writer = StringIO()
        ElementTree(root).write(writer, encoding="utf-8")
        return writer.getvalue()

    def addChange(self, variable, value):
        self._changes[variable] = str(value)
        if not self._updates_locked:
            self._send_changes()

    def _send_changes(self):
        if len(self._changes) == 0:
            return
        self.fire(Notification({ "LastChange": self.LastChange() }),
                  self.notification_channel)
        self._changes.clear()
        Timer(0.2, Event.create("UnlockUpdates"), self).register(self)
        self._updates_locked = True

    @handler("unlock_updates")
    def _on_unlock_updates(self, *args):
        self._updates_locked = False
        self._send_changes()


class RenderingController(UPnPCombinedEventsServiceController):
    
    def __init__(self, adapter, device_path, service, service_id):
        super(RenderingController, self).__init__\
            (adapter, device_path, service, service_id, UPNP_RCS_EVENT_NS)
        self._provider = adapter.provider
        self._target = None
        @handler("provider_updated", channel=self._provider.channel)
        def _on_provider_updated_handler(self, provider, changed):
            if provider != self._provider:
                return
            self._map_changes(changed)
        self.addHandler(_on_provider_updated_handler)

    def _map_changes(self, changed):
        for name, value in changed.items():
            if name == "volume":
                self.addChange("Volume", str(int(value*100)))
                continue
        
    @upnp_service
    def GetVolume(self, **kwargs):
        self.fire(Log(logging.DEBUG, "GetVolume called"), "logger")
        return [("CurrentVolume", str(int(self._provider.volume * 100)))]

    @upnp_service
    def SetVolume(self, **kwargs):
        self.fire(Log(logging.DEBUG, 'SetVolume to '
                      + kwargs["DesiredVolume"]), "logger")
        self.fire(Event.create("SetVolume", 
                               int(kwargs["DesiredVolume"]) / 100.0),
                  self.parent.provider.channel)
        return []


class ConnectionManagerController(UPnPServiceController):
    
    def __init__(self, adapter, device_path, service, service_id):
        super(ConnectionManagerController, self).__init__\
            (adapter, device_path, service, service_id)
        self._target = None

    @upnp_service
    def GetProtocolInfo(self, **kwargs):
        self.fire(Log(logging.DEBUG, "GetProtocolInfo called"), "logger")
        return [("Source", ""),
                ("Sink", "http-get:*:audio/mpeg:*")]


class AVTransportController(UPnPCombinedEventsServiceController):
    
    def __init__(self, adapter, device_path, service, service_id):
        super(AVTransportController, self).__init__\
            (adapter, device_path, service, service_id, UPNP_AVT_EVENT_NS)
        self._provider = adapter.provider
        self._target = None
        self._transport_state = "STOPPED"
        @handler("provider_updated", channel=self._provider.channel)
        def _on_provider_updated_handler(self, provider, changed):
            if provider != self._provider:
                return
            self._map_changes(changed)
        self.addHandler(_on_provider_updated_handler)
        # @handler("end_of_media", channel=self._provider.channel)

    def _format_duration(self, duration):
        return "%d:%02d:%02d" % (int(duration / 3600), 
                                 int(int(duration) % 3600 / 60),
                                 int(duration) % 60)

    def _map_changes(self, changed):
        for name, value in changed.items():
            if name == "source":
                self.addChange("AVTransportURI", value)
                self.addChange("CurrentTrackURI", value)
                continue
            if name == "source_meta_data":
                self.addChange("AVTransportURIMetaData", value)
                self.addChange("CurrentTrackMetaData", value)
                continue
            if name == "next_source":
                self.addChange("NextAVTransportURI", value)
                continue
            if name == "next_source_meta_data":
                self.addChange("NextAVTransportURIMetaData", value)
                continue
            if name == "current_track_duration":
                self.addChange("CurrentTrackDuration", 
                               "NOT_IMPLEMENTED" if value is None \
                               else self._format_duration(value))
                continue
            if name == "state":
                if value == "PLAYING":
                    self._transport_state = "PLAYING"
                    self.addChange("TransportState", self._transport_state)
                elif value == "IDLE":
                    self._transport_state = "STOPPED"
                    self.addChange("TransportState", self._transport_state)
                elif value == "PAUSED":
                    self._transport_state = "PAUSED_PLAYBACK"
                    self.addChange("TransportState", self._transport_state)
                elif value == "TRANSITIONING":
                    self._transport_state = "TRANSITIONING"
                    self.addChange("TransportState", self._transport_state)
                continue
        
    @upnp_service
    def GetTransportInfo(self, **kwargs):
        self.fire(Log(logging.DEBUG, "GetTransportInfo called"), "logger")
        return [("CurrentTransportState", self._transport_state),
                ("CurrentTransportStatus", "OK"),
                ("CurrentSpeed", "1")]

    @upnp_service
    def GetMediaInfo(self, **kwargs):
        self.fire(Log(logging.DEBUG, "GetMediaInfo called"), "logger")
        return [("NrTracks", self._provider.tracks),
                ("MediaDuration", "0:00:00"),
                ("CurrentURI", self._provider.source),
                ("CurrentURIMetaData", "NOT_IMPLEMENTED" \
                 if self._provider.source_meta_data is None \
                 else self._provider.source_meta_data),
                ("NextURI", self._provider.next_source),
                ("NextURIMetaData", self._provider.next_source_meta_data),
                ("PlayMedium", "NONE"),
                ("RecordMedium", "NOT_IMPLEMENTED"),
                ("WriteStatus", "NOT_IMPLEMENTED")]

    @upnp_service
    def GetPositionInfo(self, **kwargs):
        rel_pos = self._provider.current_position()
        self.fire(Log(logging.DEBUG, "GetPositionInfo called"), "logger")
        info = [("Track", self._provider.current_track),
                ("TrackDuration", "NOT_IMPLEMENTED" \
                 if self._provider.current_track_duration is None \
                 else self._format_duration \
                    (self._provider.current_track_duration)),
                ("TrackMetaData", "NOT_IMPLEMENTED" \
                 if self._provider.source_meta_data is None \
                 else self._provider.source_meta_data),
                ("TrackURI", self._provider.source),
                ("RelTime", "NOT_IMPLEMENTED" if rel_pos is None \
                 else self._format_duration(rel_pos)),
                ("AbsTime", "NOT_IMPLEMENTED"),
                ("RelCount", 2147483647),
                ("AbsCount", 2147483647)]
        return info

    @upnp_service
    def SetAVTransportURI(self, **kwargs):
        self.fire(Log(logging.DEBUG, 'AV Transport URI set to '
                      + kwargs["CurrentURI"]), "logger")
        self.fire(Event.create("Load", kwargs["CurrentURI"], 
                               kwargs["CurrentURIMetaData"]),
                  self.parent.provider.channel)
        return []
    
    @upnp_service
    def SetNextAVTransportURI(self, **kwargs):
        self.fire(Log(logging.DEBUG, 'Next AV Transport URI set to '
                      + kwargs["NextURI"]), "logger")
        self.fire(Event.create("PrepareNext", kwargs["NextURI"], 
                               kwargs["NextURIMetaData"]),
                  self.parent.provider.channel)
        return []
    
    @upnp_service
    def Play(self, **kwargs):
        self.fire(Log(logging.DEBUG, "Play called"), "logger")
        self.fire(Event.create("Play"),
                  self.parent.provider.channel)
        return []
    
    @upnp_service
    def Pause(self, **kwargs):
        self.fire(Log(logging.DEBUG, "Pause called"), "logger")
        self.fire(Event.create("Pause"),
                  self.parent.provider.channel)
        return []
    
    @upnp_service
    def Stop(self, **kwargs):
        self.fire(Log(logging.DEBUG, "Stop called"), "logger")
        self.fire(Event.create("Stop"),
                  self.parent.provider.channel)
        return []
    
    @upnp_service
    def Seek(self, **kwargs):
        if not (self._transport_state == "PLAYING" 
                or self._transport_state == "STOPPED"):
            raise UPnPServiceError(701) 
        unit = kwargs["Unit"]
        if unit != "REL_TIME":
            self.fire(Log(logging.DEBUG, "Seek called"), "logger")
            raise UPnPServiceError(710)
        target = kwargs["Target"]
        self.fire(Log(logging.DEBUG, "Seek to " + target + " called"), "logger")
        target = target.split(":")
        target = int(target[0]) * 3600 + int(target[1]) * 60 + int(target[2])
        self.fire(Event.create("Seek", target),
                  self.parent.provider.channel)
        return []
    
    



#
# soaplib - Copyright (C) cocy.soaplib contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#
# Adapted to standard etree implementation (mnl at mnl.de)

from xml import etree

import cocy.soaplib

from cocy.soaplib.serializers import Base
from cocy.soaplib.serializers import nillable_element
from cocy.soaplib.serializers import nillable_value

from cocy.soaplib.util.odict import odict as TypeInfo

class ClassSerializerMeta(type(Base)):
    '''
    This is the metaclass that populates ClassSerializer instances with
    the appropriate datatypes for (de)serialization.
    '''

    def __new__(cls, cls_name, cls_bases, cls_dict):
        '''
        This initializes the class, and sets all the appropriate types onto the
        class for serialization.
        '''

        type_name = cls_dict.get("__type_name__", None)
        if type_name is None:
            cls_dict["__type_name__"] = cls_name

        # get base class (if exists) and enforce single inheritance
        extends = cls_dict.get("__extends__", None)

        if extends is None:
            for b in cls_bases:
                base_types = getattr(b, "_type_info", None)

                if not (base_types is None):
                    if not (extends is None or cls_dict["__extends__"] is b):
                        raise Exception("WSDL 1.1 does not support multiple "
                                        "inheritance")

                    try:
                        if len(base_types) > 0 and issubclass(b, Base):
                            cls_dict["__extends__"] = extends = b
                    except:
                        print extends
                        raise

        # populate soap members
        if not ('_type_info' in cls_dict):
            cls_dict['_type_info'] = _type_info = TypeInfo()

            for k,v in cls_dict.items():
                if not k.startswith('__'):
                    subc = False
                    try:
                        if issubclass(v,Base):
                            subc = True
                    except:
                        pass

                    if subc:
                        _type_info[k] = v
                        if issubclass(v, Array) and v.serializer is None:
                            raise Exception("%s.%s is an array of what?" %
                                                                  (cls_name, k))
        else:
            _type_info = cls_dict['_type_info']
            if not isinstance(_type_info, TypeInfo):
                cls_dict['_type_info'] = TypeInfo(_type_info)

        return type.__new__(cls, cls_name, cls_bases, cls_dict)

class NonExtendingClass(object):
    def __setattr__(self, k, v):
        if not hasattr(self,k) and getattr(self, '__NO_EXTENSION', False):
            raise Exception("'%s' object is not extendable at this point in "
                            "code.\nInvalid member '%s'" %
                                                  (self.__class__.__name__, k) )
        object.__setattr__(self,k,v)

class ClassSerializerBase(NonExtendingClass, Base):
    """
    If you want to make a better class serializer, this is what you should
    inherit from
    """

    def __init__(self, **kwargs):
        super(ClassSerializerBase,self).__init__()
        self.__reset_members(self.__class__, kwargs)

    def __reset_members(self, cls, kwargs):
        extends = getattr(cls, "__extends__", None)
        if not (extends is None):
            self.__reset_members(extends, kwargs)

        for k in cls._type_info.keys():
            setattr(self, k, kwargs.get(k, None))

    def __len__(self):
        return len(self._type_info)

    def __getitem__(self,i):
        return getattr(self, self._type_info.keys()[i], None)

    @classmethod
    def get_serialization_instance(cls, value):
        # if the instance is a list, convert it to a cls instance.
        # this is only useful when deserializing descriptor.in_message as it's
        # the only time when the member order is not arbitrary (as the members
        # are declared and passed around as sequences of arguments, unlike
        # dictionaries in a regular class definition).
        if isinstance(value, list) or isinstance(value, tuple):
            assert len(value) <= len(cls._type_info)

            inst = cls()

            keys = cls._type_info.keys()
            for i in range(len(value)):
                setattr(inst, keys[i], value[i])

        elif isinstance(value, dict):
            inst = cls()

            for k in cls._type_info:
                setattr(inst, k, value.get(k,None))

        else:
            inst = value

        return inst

    @classmethod
    def get_deserialization_instance(cls):
        return cls()

    @classmethod
    def get_members(cls, inst, parent):
        parent_cls = getattr(cls, '__extends__', None)
        if not (parent_cls is None):
            parent_cls.get_members(inst, parent)

        for k, v in cls._type_info.items():
            mo = v.Attributes.max_occurs
            subvalue = getattr(inst, k, None)

            if mo == 'unbounded' or mo > 1:
                if subvalue != None:
                    for sv in subvalue:
                        v.to_xml(sv, cls.get_namespace(), parent, k)

            else:
                v.to_xml(subvalue, cls.get_namespace(), parent, k)

    @classmethod
    @nillable_value
    def to_xml(cls, value, tns, parent_elt, name=None):
        if name is None:
            name = cls.get_type_name()

        element = etree.ElementTree.SubElement(parent_elt, "{%s}%s" % (tns, name))

        inst = cls.get_serialization_instance(value)

        cls.get_members(inst, element)

    @classmethod
    @nillable_element
    def from_xml(cls, element):
        inst = cls.get_deserialization_instance()

        for c in element:
            if isinstance(c, etree.ElementTree.Comment):
                continue

            key = c.tag.split('}')[-1]

            member = cls._type_info.get(key, None)
            clz = getattr(cls,'__extends__', None)
            while not (clz is None) and (member is None):
                member = clz._type_info.get(key, None)
                clz = getattr(clz,'__extends__', None)

            if member is None:
                raise Exception('the %s object does not have a "%s" member' %
                                                             (cls.__name__,key))

            mo = member.Attributes.max_occurs
            if mo == 'unbounded' or mo > 1:
                value = getattr(inst, key, None)
                if value is None:
                    value = []

                value.append(member.from_xml(c))
                setattr(inst, key, value)

            else:
                setattr(inst, key, member.from_xml(c))

        return inst

    @staticmethod
    def resolve_namespace(cls, default_ns):
        if getattr(cls, '__extends__', None) != None:
            cls.__extends__.resolve_namespace(cls.__extends__, default_ns)

        Base.resolve_namespace(cls, default_ns)

        for k, v in cls._type_info.items():
            if v.__type_name__ is Base.Empty:
                v.__namespace__ = cls.get_namespace()
                v.__type_name__ = "%s_%sType" % (cls.get_type_name(), k)

            if v != cls:
                v.resolve_namespace(v, default_ns)

    @classmethod
    def add_to_schema(cls, schema_entries):
        if cls.get_type_name() is Base.Empty:
            (child,) = cls._type_info.values()
            cls.__type_name__ = '%sArray' % child.get_type_name()

        if not schema_entries.has_class(cls):
            if not (getattr(cls, '__extends__', None) is None):
                cls.__extends__.add_to_schema(schema_entries)

            complex_type = etree.ElementTree.Element("{%s}complexType" % cocy.soaplib.ns_xsd)
            complex_type.set('name',cls.get_type_name())

            sequence_parent = complex_type
            if not (getattr(cls, '__extends__', None) is None):
                cls.__extends__.add_to_schema(schema_entries)

                complex_content = etree.ElementTree.SubElement(complex_type,
                                          "{%s}complexContent" % cocy.soaplib.ns_xsd)
                extension = etree.ElementTree.SubElement(complex_content, "{%s}extension"
                                                               % cocy.soaplib.ns_xsd)
                extension.set('base', cls.__extends__.get_type_name_ns())
                sequence_parent = extension

            sequence = etree.ElementTree.SubElement(sequence_parent, '{%s}sequence' %
                                                                cocy.soaplib.ns_xsd)

            for k, v in cls._type_info.items():
                if v != cls:
                    v.add_to_schema(schema_entries)

                member = etree.ElementTree.SubElement(sequence, '{%s}element' %
                                                                cocy.soaplib.ns_xsd)
                member.set('name', k)
                member.set('type', v.get_type_name_ns())

                if v.Attributes.min_occurs != 1: # 1 is the xml schema default
                    member.set('minOccurs', str(v.Attributes.min_occurs))
                if v.Attributes.max_occurs != 1: # 1 is the xml schema default
                    member.set('maxOccurs', str(v.Attributes.max_occurs))

                # True is the xml schema default
                if bool(v.Attributes.nillable) == True:
                    member.set('nillable', 'true')

            schema_entries.add_complex_type(cls, complex_type)

            # simple node
            element = etree.ElementTree.Element('{%s}element' % cocy.soaplib.ns_xsd)
            element.set('name',cls.get_type_name())
            element.set('type',cls.get_type_name_ns())

            schema_entries.add_element(cls, element)

    @staticmethod
    def produce(namespace, type_name, members):
        """
        Lets you create a class programmatically.
        """

        cls_dict = {}

        cls_dict['__namespace__'] = namespace
        cls_dict['__type_name__'] = type_name
        cls_dict['_type_info'] = TypeInfo(members)

        return ClassSerializerMeta(type_name, (ClassSerializer,), cls_dict)

class ClassSerializer(ClassSerializerBase):
    """
    The general complexType factory. The __call__ method of this class will
    return instances, contrary to primivites where the same call will result in
    customized duplicates of the original class definition.
    Those who'd like to customize the class should use the customize method.
    (see cocy.soaplib.serializers.base.Base)
    """

    __metaclass__ = ClassSerializerMeta

class Array(ClassSerializer):
    def __new__(cls, serializer, ** kwargs):
        retval = cls.customize(**kwargs)

        # hack to default to unbounded arrays when the user didn't specify
        # max_occurs.
        if serializer.Attributes.max_occurs == 1:
            serializer = serializer.customize(max_occurs='unbounded')

        if serializer.get_type_name() is Base.Empty:
            member_name = serializer.__base_type__.get_type_name()
            if cls.__type_name__ is None:
                cls.__type_name__ = Base.Empty # to be resolved later

        else:
            member_name = serializer.get_type_name()
            if cls.__type_name__ is None:
                cls.__type_name__ = '%sArray' % serializer.get_type_name()

        retval.__type_name__ = '%sArray' % member_name
        retval._type_info = {member_name: serializer}

        return retval

    # the array belongs to its child's namespace, it doesn't have its own
    # namespace.
    @staticmethod
    def resolve_namespace(cls, default_ns):
        (serializer,) = cls._type_info.values()

        serializer.resolve_namespace(serializer, default_ns)

        if cls.__namespace__ is None:
            cls.__namespace__ = serializer.get_namespace()

        if cls.__namespace__ in cocy.soaplib.const_prefmap:
            cls.__namespace__ = default_ns

        ClassSerializer.resolve_namespace(cls, default_ns)

    @classmethod
    def get_serialization_instance(cls, value):
        inst = ClassSerializer.__new__(Array)

        (member_name,) = cls._type_info.keys()
        setattr(inst, member_name, value)

        return inst

    @classmethod
    @nillable_element
    def from_xml(cls, element):
        retval = []
        (serializer,) = cls._type_info.values()

        for child in element.getchildren():
            retval.append(serializer.from_xml(child))

        return retval



"""
..
   This file is part of the CoCy program.
   Copyright (C) 2012 Michael N. Lipp
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

.. codeauthor:: mnl
"""
from circuits.core.components import BaseComponent
from circuits.core.handlers import handler
from cocy.upnp.ssdp import SSDPTranceiver, UPnPSearchRequest, UPnPDeviceByeBye
from circuits.core.utils import findroot, flatten
from cocy.upnp import UPNP_ROOTDEVICE, SSDP_DEVICE_SCHEMA
from circuits_bricks.web import Client
from circuits_bricks.core.timers import Timer
from circuits.web.client import Request
import httplib
from xml.etree.ElementTree import XML
from urlparse import urljoin
from copy import copy

class UPnPDeviceDirectory(BaseComponent):

    channel = "upnp"

    def __init__(self, *args, **kwargs):
        super(UPnPDeviceDirectory, self).__init__(*args, **kwargs)

    @handler("started", channel="application")
    def _on_started(self, component):
        self.fire(UPnPSearchRequest(), "ssdp")

    @handler("upnp_device_alive", channel="*")
    def _on_device_alive \
        (self, location, notification_type, max_age, server, usn):
        if notification_type == UPNP_ROOTDEVICE:
            if not any(map(lambda c: isinstance(c, UPnPRootDevice) \
                           and c.usn == usn, self.components.copy())):
                UPnPRootDevice(location, max_age, usn).register(self)

    def register(self, parent):
        super(UPnPDeviceDirectory, self).register(parent)
        # SSDP transceiver, may exist only once
        if not any([isinstance(c, SSDPTranceiver) \
                    for c in flatten(findroot(self))]):
            SSDPTranceiver().register(self.parent)
        return self

    @property
    def devices(self):
        return filter(lambda c: isinstance(c, UPnPRootDevice) \
                      and c.ready, self.components.copy())


class IconInfo(object):
    def __init__(self, width, height, url):
        self.width = width
        self.height = height
        self.url = url


class UPnPRootDevice(BaseComponent):

    def __init__(self, location, max_age, usn):
        super(UPnPRootDevice, self).__init__(channel=usn)
        self._location = location
        self._usn = usn
        self._ready = False
        self._comm_chan = "client." + usn
        self._client = Client(location, channel=self._comm_chan).register(self)
        @handler("response", channel=self._comm_chan)
        def _on_response(self, response):
            if response.status == httplib.OK:
                self._initialize(response.read())
        self.addHandler(_on_response)
        @handler("error", channel=self._comm_chan)
        def _on_error(self, *args, **kwargs):
            self._client.close()
            self.unregister()
        self.addHandler(_on_error)
        self.fire(Request("GET", self._location), self._client)
        self._expiry_timer \
            = Timer(max_age, UPnPDeviceByeBye(usn)).register(self)

    def _initialize(self, xml_src):
        data = XML(xml_src)
        self._friendly_name = data.findtext \
            ("{%s}device/{%s}friendlyName" \
             % (SSDP_DEVICE_SCHEMA, SSDP_DEVICE_SCHEMA))
        icons = data.findall \
            ("{%s}device/{%s}iconList/{%s}icon" \
             % (SSDP_DEVICE_SCHEMA, SSDP_DEVICE_SCHEMA, SSDP_DEVICE_SCHEMA))
        self._icons = []
        for icon in icons:
            width = int(icon.findtext("{%s}width" % SSDP_DEVICE_SCHEMA))
            height = int(icon.findtext("{%s}height" % SSDP_DEVICE_SCHEMA))
            url = urljoin(self._location,
                          icon.findtext("{%s}url" % SSDP_DEVICE_SCHEMA))
            self._icons.append(IconInfo(width, height, url))
        self._ready = True

    @handler("upnp_device_alive")
    def _on_device_alive \
        (self, location, notification_type, max_age, server, usn):
        self._expiry_timer.interval = max_age
        self._expiry_timer.reset()

    @handler("upnp_device_bye_bye")
    def _on_device_bye_bye (self, usn):
        self._client.close()
        self.unregister()

    @property
    def usn(self):
        return getattr(self, "_usn", None)
        
    @property
    def location(self):
        return getattr(self, "_location", None)

    @property
    def ready(self):
        return getattr(self, "_ready", None)

    @property
    def friendly_name(self):
        return getattr(self, "_friendly_name", None)

    @property
    def icons(self):
        return copy(getattr(self, "_icons", None))
    
    @property
    def valid_until(self):
        return self._expiry_timer._eTime \
            if hasattr(self, "_expiry_timer") else None
        


"""
..
   This file is part of the CoCy program.
   Copyright (C) 2011 Michael N. Lipp
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from cocy.upnp.adapters.adapter import upnp_service, UPnPServiceController,\
    upnp_state

class BinarySwitchPowerController(UPnPServiceController):
    
    def __init__(self, adapter, device_path, service, service_id):
        super(BinarySwitchPowerController, self).__init__\
            (adapter, device_path, service, service_id)
        self._target = None

    @upnp_state
    def Target(self):
        return self._target

    @upnp_state(evented_by="state")
    def Status(self):
        return self.parent.provider.state

    @upnp_service
    def SetTarget(self, **kwargs):
        self._target \
            = (kwargs["newTargetValue"].lower() in ["1", "yes", "true"])
        self.parent.provider.state = self._target 
        return []

    @upnp_service
    def GetTarget(self, **kwargs):
        return [("RetTargetValue", "1" if self.parent.provider.state else "0")]

    @upnp_service
    def GetStatus(self, **kwargs):
        return [("ResultStatus", "1" if self.parent.provider.state else "0")]



"""
..
   This file is part of the CoCy program.
   Copyright (C) 2011 Michael N. Lipp
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

.. codeauthor:: mnl
"""
from circuits.core.components import BaseComponent
from circuits.core.handlers import handler
from circuits_bricks.net.sockets import UDPMCastServer
import os
from circuits.io.events import Write
import platform
from socket import gethostname, gethostbyname
import time
from circuits_bricks.core.timers import Timer
from cocy.upnp import SSDP_ADDR, SSDP_PORT, SSDP_SCHEMAS, UPNP_ROOTDEVICE,\
    SERVER_HELLO
from circuits_bricks.misc import ComponentQuery
from circuits.core.events import Event
from circuits.web.controllers import Controller
from email.utils import formatdate


class SSDPTranceiver(BaseComponent):
    '''The SSDP protocol server component
    '''

    channel = "ssdp"

    def __init__(self, **kwargs):
        '''
        Constructor
        '''
        kwargs.setdefault("channel", self.channel)
        super(SSDPTranceiver, self).__init__(**kwargs)

        # The underlying network connection, used by both the sender
        # and the receiver 
        self._server = UDPMCastServer((SSDP_ADDR, SSDP_PORT),
                                     **kwargs).register(self)
        self._server.setTTL(2)

        # Our associated SSDP message sender
        SSDPSender().register(self)
        
        # Our associated SSDP message receiver
        SSDPReceiver().register(self)



class SSDPSender(BaseComponent):
    '''The SSDP Protocol sender component
    '''

    channel = "ssdp"
    _template_dir = os.path.join(os.path.dirname(__file__), "templates")
    _template_cache = {}
    _message_env = {}
    _message_expiry = 1800
    _boot_id = int(time.time())
    _timers = dict()

    def __init__(self, channel=channel):
        '''
        Constructor
        '''
        super(SSDPSender, self).__init__(channel=channel)

        # Setup the common entries in the dictionary that will be usd to
        # fill the UPnP templates.
        self._message_env['BOOTID'] = self._boot_id
        self._message_env['SERVER'] = SERVER_HELLO
        self.hostaddr = gethostbyname(gethostname())
        if self.hostaddr.startswith("127.") and not "." in gethostname():
            try:
                self.hostaddr = gethostbyname(gethostname() + ".")
            except:
                pass

    @handler("config_value", channel="configuration")
    def _on_config_value(self, section, option, value):
        if not section == "upnp":
            return
        if option == "max-age":
            self._message_expiry = int(value)

    @handler("mgmt_controller_query")
    def _on_controller_query(self):
        return Controller()

    @handler("device_available", channel="upnp")
    def _on_device_available(self, event, upnp_device):
        self._update_message_env(upnp_device)
        self._send_device_messages(upnp_device, "available")
        # Service announcements
        for service in upnp_device.services:
            self._send_service_message(upnp_device, service, "available")
        # Handle repeats
        if getattr(event, 'times_sent', 0) < 3:
            self._timers[upnp_device.uuid] \
                = Timer(0.25, event, *event.channels).register(self)
            event.times_sent = getattr(event, 'times_sent', 0) + 1
        else:
            self._timers[upnp_device.uuid] \
                = Timer(self._message_expiry / 4,
                        event, *event.channels).register(self)
   
    @handler("device_unavailable", channel="upnp")
    def _on_device_unavailable(self, event, upnp_device):
        if self._timers.has_key(upnp_device.uuid):
            self._timers[upnp_device.uuid].unregister()
            del self._timers[upnp_device.uuid]
        self._update_message_env(upnp_device)
        self._send_device_messages(upnp_device, "unavailable")
        # Service announcements
        for service in upnp_device.services:
            self._send_service_message(upnp_device, service, "unavailable")
   
    @handler("device_match")
    def _on_device_match(self, upnp_device, inquirer, search_target):
        self._update_message_env(upnp_device)
        if search_target == "ssdp:all":
            self._send_device_messages(upnp_device, "result", inquirer)
        else:
            if search_target == "upnp:rootdevice":
                self._send_root_message(upnp_device, "notify-result", inquirer)
            elif search_target.startswith("uuid:"):
                self._send_uuid_message(upnp_device, "notify-result", inquirer)
            elif search_target.startswith(SSDP_SCHEMAS + ":device:"):
                self._send_device_message \
                    (upnp_device, "notify-result", inquirer)
            elif search_target.startswith(SSDP_SCHEMAS + ":service:"):
                for service in upnp_device.services:
                    if not search_target.startswith \
                        (SSDP_SCHEMAS + ":service:" + service.type + ":"):
                        continue
                    self._send_service_message \
                        (upnp_device, service, "result", inquirer)
            
    @handler("upnp_search_request")
    def _on_search_request(self, event, search_target=UPNP_ROOTDEVICE, mx=1):
        self._send_template("m-search-request", 
                            { "ST": search_target, "MX": mx })
        # Handle repeats
        if getattr(event, 'times_sent', 0) < 3:
            Timer(mx, event, *event.channels).register(self)
            event.times_sent = getattr(event, 'times_sent', 0) + 1
            
    def _update_message_env(self, upnp_device):
        self._message_env['CACHE-CONTROL'] = self._message_expiry
        self._message_env['CONFIGID'] = upnp_device.config_id
        self._message_env['DATE'] = formatdate(usegmt=True)
        self._message_env['LOCATION'] = "http://" + self.hostaddr + ":" \
            + str(upnp_device.web_server_port) + "/" + upnp_device.uuid \
            + "/description.xml"
        
    def _send_device_messages(self, upnp_device, msg_type, 
                              to=(SSDP_ADDR, SSDP_PORT)):
        template = "notify-%s" % msg_type
        # There is an extra announcement for root devices
        if upnp_device.root_device:
            self._send_root_message(upnp_device, template, to)
        # Device UUID announcement
        self._send_uuid_message(upnp_device, template, to)
        # Device type announcement
        self._send_device_message(upnp_device, template, to)
        
    def _send_root_message(self, upnp_device, template, 
                           to=(SSDP_ADDR, SSDP_PORT)):
        self._message_env['NT'] = 'upnp:rootdevice'
        self._message_env['USN'] = 'uuid:' + upnp_device.uuid \
            + '::upnp:rootdevice'
        self._send_template(template, self._message_env, to)

    def _send_uuid_message(self, upnp_device, template, 
                           to=(SSDP_ADDR, SSDP_PORT)):
        self._message_env['NT'] = 'uuid:' + upnp_device.uuid
        self._message_env['USN'] = 'uuid:' + upnp_device.uuid
        self._send_template(template, self._message_env, to)

    def _send_device_message(self, upnp_device, template, 
                           to=(SSDP_ADDR, SSDP_PORT)):
        self._message_env['NT'] = SSDP_SCHEMAS + ":device:" \
            + upnp_device.type_ver
        self._message_env['USN'] = 'uuid:' + upnp_device.uuid \
            + "::" + self._message_env['NT']
        self._send_template(template, self._message_env, to)

    def _send_service_message(self, upnp_device, service, msg_type, \
                              to=(SSDP_ADDR, SSDP_PORT)):
        template = "notify-%s" % msg_type
        self._message_env['NT'] = SSDP_SCHEMAS + ":service:" \
            + service.type_ver
        self._message_env['USN'] = 'uuid:' + upnp_device.uuid \
            + "::" + self._message_env['NT']
        self._send_template(template, self._message_env, to)
        
    def _send_template(self, template_name, data, to=(SSDP_ADDR, SSDP_PORT)):
        template = self._get_template(template_name)
        message = template % data
        headers = ""
        for line in message.splitlines():
            headers = headers + line + "\r\n"
        headers += "\r\n"
        self.fireEvent(Write(to, headers))
                    
    def _get_template(self, name):
        if self._template_cache.has_key(name):
            return self._template_cache[name]
        template_file = open(os.path.join(self._template_dir, name))
        template = template_file.read()
        self._template_cache[name] = template
        return template


class UPnPDeviceMatch(Event):
    name = "device_match"
    
    def __init__(self, component, inquirer, search_target):
        super(UPnPDeviceMatch, self)\
            .__init__(component, inquirer, search_target)


class UPnPDeviceQuery(ComponentQuery):
    
    def __init__(self, query_function, inquirer, search_target):
        super(UPnPDeviceQuery, self).__init__(query_function)
        self.name = super(UPnPDeviceQuery, self).name
        self._inquirer = inquirer
        self._search_target = search_target

    def decide(self, component):
        res = super(UPnPDeviceQuery, self).decide(component)
        if res != None:
            component.fire(UPnPDeviceMatch(component, self._inquirer, \
                                           self._search_target), "ssdp")

class UPnPDeviceAlive(Event):
    
    name = "upnp_device_alive"
    
    def __init__(self, location, notification_type, max_age, server, usn):
        super(UPnPDeviceAlive, self).__init__ \
            (location, notification_type, max_age, server, usn)
        self.channels = (usn,)


class UPnPDeviceByeBye(Event):
    
    name = "upnp_device_bye_bye"
    
    def __init__(self, usn):
        super(UPnPDeviceByeBye, self).__init__(usn)
        self.channels = (usn,)


class UPnPSearchRequest(Event):
    
    name = "upnp_search_request"
    
    def __init__(self, search_target=UPNP_ROOTDEVICE, mx=1, **kwargs):
        super(UPnPSearchRequest, self).__init__(search_target, mx, **kwargs)


class SSDPReceiver(BaseComponent):

    channel = "ssdp"

    def __init__(self, channel = channel):
        super(SSDPReceiver, self).__init__(channel=channel)

    @handler("read")
    def _on_read(self, address, data):
        
        def istartswith(line, prefix):
            return line[0:len(prefix)].upper() == prefix

        def parse_lines(lines):
            ''' Helper function that parses the information and puts it in an
                anonymous type.'''
            res = type("", (), {})()
            for line in lines:
                if istartswith(line, "CACHE-CONTROL:"):
                    s = line.split(":", 1)[1].strip()
                    setattr(res, "max_age", int(s.split("=", 1)[1].strip()))
                elif istartswith(line, "LOCATION:"):
                    setattr(res, "location", line.split(":", 1)[1].strip())
                elif istartswith(line, "NT:"):
                    setattr(res, "notification_type", 
                            line.split(":", 1)[1].strip())
                elif istartswith(line, "ST:"):
                    # Handle search responses like alive messages
                    setattr(res, "notification_type", 
                            line.split(":", 1)[1].strip())
                elif istartswith(line, "NTS:"):
                    setattr(res, "sub_type", line.split(":", 1)[1].strip())
                elif istartswith(line, "SERVER:"):
                    setattr(res, "server", line.split(":", 1)[1].strip())
                elif istartswith(line, "USN:"):
                    setattr(res, "usn", line.split(":", 1)[1].strip())
            return res
        
        lines = data.splitlines()
        if istartswith(lines[0], "M-SEARCH "):
            # This is a search. It's up to us to repond if we have
            # matching devices
            search_target = None
            for line in lines[1:len(lines)-1]:
                if not line.upper().startswith("ST:"):
                    continue
                search_target = line.split(':', 1)[1].strip()
                f = lambda dev: True
                # TODO: add criteria
                if search_target == "upnp:rootdevice":
                    f = lambda dev: dev.root_device
                self.fire(UPnPDeviceQuery(f, address, search_target), "upnp")                        
                return
        elif istartswith(lines[0], "NOTIFY "):
            # A status change (or confirmation notification. Translate into
            # an event to inform however is interested in this.
            data = parse_lines(lines[1:])
            if data.sub_type == "ssdp:alive":
                self.fire(UPnPDeviceAlive\
                          (data.location, data.notification_type, 
                           data.max_age, data.server, data.usn))
            elif data.sub_type == "ssdp:byebye":
                self.fire(UPnPDeviceByeBye(data.usn))
        elif istartswith(lines[0], "HTTP/1.1 200 OK"):
            # A response to our own M-SEARCH. This is handled like a 
            # status change/confirmation (can only be an alive notification,
            # of course).
            data = parse_lines(lines[1:])
            self.fire(UPnPDeviceAlive\
                      (data.location, data.notification_type, 
                       data.max_age, data.server, data.usn))



"""
..
   This file is part of the CoCy program.
   Copyright (C) 2011 Michael N. Lipp
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from cocy.upnp.adapters.adapter import UPnPDeviceAdapter
from cocy.providers import BinarySwitch, MediaPlayer
from cocy.upnp import UPNP_SERVICE_ID_PREFIX
from cocy.upnp.adapters.home_automation import BinarySwitchPowerController
from cocy.upnp.adapters.audio_video import RenderingController,\
    ConnectionManagerController, AVTransportController

UPnPDeviceAdapter.add_adapter \
    (BinarySwitch, UPnPDeviceAdapter.DeviceProperties \
     ("BinaryLight", 0.9, 1, 1,
      [("SwitchPower:1", "SwitchPower", BinarySwitchPowerController)],
      "_common_device_desc"))

UPnPDeviceAdapter.add_adapter \
    (MediaPlayer, UPnPDeviceAdapter.DeviceProperties \
     ("MediaRenderer", 1.0, 1, 1,
      [("RenderingControl:1.0", "RenderingControl", RenderingController),
       ("ConnectionManager:1.0", "ConnectionManager",
        ConnectionManagerController),
       ("AVTransport:1.0", "AVTransport", AVTransportController)],
      "_common_device_desc"))

#UPnPDeviceAdapter.add_adapter \
#    (MediaPlayer, UPnPDeviceAdapter.DeviceProperties \
#     ("MediaRenderer", 2, 1, 1,
#      [("RenderingControl:2", "RenderingControl", RenderingController),
#       ("ConnectionManager:2", "ConnectionManager",
#        ConnectionManagerController),
#       ("AVTransport:2", "AVTransport", AVTransportController)],
#      "_common_device_desc"))



"""
..
   This file is part of the CoCy program.
   Copyright (C) 2011 Michael N. Lipp
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

.. codeauthor:: mnl
"""
from circuits.core.components import BaseComponent
from circuits.core.handlers import handler
from circuits.web.servers import BaseServer
from circuits_bricks.web import ScopeDispatcher, ScopedChannel
from circuits.web.controllers import BaseController, expose
from circuits.core.events import Event
from circuits.core.utils import findroot, flatten
from cocy.upnp.service import UPnPService
from cocy.upnp.ssdp import SSDPTranceiver
from cocy.providers import Provider
import anydbm
import os
from xml.etree.ElementTree import Element, QName, SubElement
from cocy.upnp import UPNP_CONTROL_NS
from circuits.web.errors import HTTPError
from cocy.soaplib import ns_soap_env
from cocy.misc import buildSoapResponse


class DeviceAvailable(Event):
    name = "device_available"
    

class DeviceUnavailable(Event):
    name = "device_unavailable"


class UPnPDeviceServer(BaseComponent):
    """
    This component keeps track of the :class:`cocy.providers.Provider` 
    instances and creates or removes the corresponding 
    :class:`cocy.upnp.device.UPnPDeviceAdapter` components.
    
    Notifications are sent when a new device is added 
    (:class:`cocy.upnp.ssdp.DeviceAvailable`) or removed
    (:class:`cocy.upnp.ssdp.DeviceUnavailable`)
    """
    channel = "upnp"
    
    def __init__(self, path, channel=channel):
        super(UPnPDeviceServer, self).__init__(channel=channel)
        self._started = False

        # Build a web (HTTP) server for handling requests. This is
        # the server that will be announced by SSDP, so it has
        # no fixed port number.
        self.web_server = BaseServer(("", 0), channel="upnp-web").register(self)
        # Dispatcher for "/upnp-web".
        disp = ScopeDispatcher(channel="upnp-web").register(self.web_server)
        # Dummy root controller prevents requests for nested resources
        # from failing.
        DummyRoot().register(disp)
        
        # Initially empty list of providers
        self._devices = []
        
        # The configuration id, incremented every time the 
        # configuration changes
        self.config_id = 1
        
        # Open the database for uuid persistence 
        self._uuid_db = anydbm.open(os.path.join(path, 'upnp_uuids'), 'c')

    def register(self, parent):
        super(UPnPDeviceServer, self).register(parent)
        # SSDP transceiver, may exist only once
        if not any([isinstance(c, SSDPTranceiver) \
                    for c in flatten(findroot(self))]):
            SSDPTranceiver().register(self.parent)
        return self

    @handler("registered", channel="*")
    def _on_registered(self, component, manager):
        if not isinstance(component, Provider):
            return
        from cocy.upnp.adapters.adapter import UPnPDeviceAdapter
        device = UPnPDeviceAdapter(self, component, self.config_id, \
                                   self._uuid_db, self.web_server.port)
        if not device.valid:
            return
        device.register(self)
        self._devices.append(device)
        if self._started:
            self.fireEvent(DeviceAvailable(device))

    @handler("unregister")
    def _on_unregister(self, component, manager):
        if not isinstance(component, Provider):
            return
        

    @handler("started", channel="application")
    def _on_started (self, component):
        self._started = True
        for device in self._devices:
            self.fireEvent(DeviceAvailable(device))

    @handler("stopped", channel="*", priority=100, filter=True)
    def _on_stopped(self, event, component):
        if not self._started:
            return
        self._started = False
        for device in self._devices:
            self.fireEvent(DeviceUnavailable(device))
        self._uuid_db.close()
        self.fireEvent(event)
        return True

    @property
    def providers(self):
        return [device.provider for device in getattr(self, "_devices", [])]


class UPnPError(HTTPError):

    _error_descs = { 401: "Invalid Action",
                     402: "Invalid Args",
                     501: "Action Failed",
                     600: "Argument Value Invalid",
                     601: "Argument Value Out of Range",
                     602: "Optional Action Not Implemented",
                     603: "Out of Memory",
                     604: "Human Intervention Required",
                     605: "String Argument Too Long"}
    
    def __init__(self, request, response, error_code, error_desc = None):
        super(UPnPError, self).__init__(request, response, 500)
        if error_desc is None:
            error_desc = self._error_descs.get(error_code, "Unknown")
        result = Element(QName(ns_soap_env, "Fault"))
        SubElement(result, QName(ns_soap_env, "faultcode")).text \
            = str(QName(ns_soap_env, "Client"))
        SubElement(result, QName(ns_soap_env, "faultstring")).text \
            = "UPnPError"
        detail = SubElement(result, "detail")
        upnp_error = SubElement(detail, QName(UPNP_CONTROL_NS, "UPnPError"))
        SubElement(upnp_error, QName(UPNP_CONTROL_NS, "errorCode")).text \
            = str(error_code)
        SubElement(upnp_error, QName(UPNP_CONTROL_NS, "errorDescription")) \
            .text = error_desc
        self._fault = buildSoapResponse(self.response, result)
    
    def __str__(self):
        return self._fault 


class DummyRoot(BaseController):
    
    channel = ScopedChannel("upnp-web", "/")
    
    @expose("f3c3G5")
    def index(self):
        return "Hello"
    




"""
..
   This file is part of the CoCy program.
   Copyright (C) 2011 Michael N. Lipp
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

.. codeauthor:: mnl
"""
from cocy.soaplib.soap import from_soap
from xml.etree.ElementTree import ElementTree, Element, SubElement, QName
import cocy.soaplib
from StringIO import StringIO

def splitQTag (tag):
    tag_ns, tag_name = tag.split("}", 1)
    tag_ns = tag_ns[1:]
    return (tag_ns, tag_name)

def parseSoapRequest(request):
    # Test if this is a SOAP request. SOAP 1.1 specifies special
    # header, SOAP 1.2 special Content-Type
    soapAction = request.headers["SOAPAction"];
    import cgi
    contentType = cgi.parse_header(request.headers["Content-Type"]);
    if (not soapAction and contentType[0] != "application/soap+xml"):
        return
    # Get body data of request
    body = request.body.read()
    # Use soaplib to separate header and payload
    charset = contentType[1].get('charset',None)
    if charset is None:
        charset = 'utf-8'
            
    payload, soapheader = from_soap(body, charset)
    from cocy.soaplib.soap import collapse_swa
    payload = collapse_swa(contentType, payload)

    if payload is not None:
        soapAction = payload.tag

    return soapAction, soapheader, payload


def buildSoapResponse(response, body):
    # construct the soap response, and serialize it
    envelope = Element('{%s}Envelope' % cocy.soaplib.ns_soap_env)
    # body
    soap_body = SubElement(envelope, '{%s}Body' % cocy.soaplib.ns_soap_env)
    soap_body.append(body)

    writer = StringIO()
    response.headers["Content-Type"] = "text/xml; charset=utf-8"
    writer.write("<?xml version='1.0' encoding='utf-8'?>")
    ElementTree(envelope).write(writer, encoding="utf-8")
    return writer.getvalue()


def set_ns_prefixes(elem, prefix_map):

    # build uri map and add to root element
    uri_map = {}
    for prefix, uri in prefix_map.items():
        uri_map[uri] = prefix
        elem.set("xmlns" + "" if not prefix else (":" + prefix), uri)

    # fixup all elements in the tree
    memo = {}
    for elem in elem.getiterator():
        _fixup_element_prefixes(elem, uri_map, memo)
        

def _fixup_element_prefixes(elem, uri_map, memo):
    def fixup(name):
        try:
            return memo[name]
        except KeyError:
            if isinstance(name, QName):
                name = str(name)
            if name[0] != "{":
                return
            uri, tag = name[1:].split("}")
            if uri in uri_map:
                new_name = (uri_map[uri] + ":") if uri_map[uri] else "" + tag
                memo[name] = new_name
                return new_name
    # fix element name
    name = fixup(elem.tag)
    if name:
        elem.tag = name
    # fix attribute names
    for key, value in elem.items():
        name = fixup(key)
        if name:
            elem.set(name, value)
            del elem.attrib[key]



#
# soaplib - Copyright (C) cocy.soaplib contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

import cocy.soaplib
from lxml import etree

from cocy.soaplib.soap import Message
from cocy.soaplib.soap import MethodDescriptor
from cocy.soaplib.serializers.clazz import TypeInfo
from cocy.soaplib.util.odict import odict

import logging
logger = logging.getLogger(__name__)

_pref_wsa = cocy.soaplib.prefmap[soaplib.ns_wsa]

def rpc(*params, **kparams):
    '''
    This is a method decorator to flag a method as a remote procedure call.  It
    will behave like a normal python method on a class, and will only behave
    differently when the keyword '_method_descriptor' is passed in, returning a
    'MethodDescriptor' object.  This decorator does none of the soap/xml
    serialization, only flags a method as a soap method.  This decorator should
    only be used on member methods of an instance of ServiceBase.
    '''

    def explain(f):
        def explain_method(*args, **kwargs):
            if '_method_descriptor' in kwargs:
                # input message
                def get_input_message(ns):
                    _in_message = kparams.get('_in_message', f.func_name)
                    _in_variable_names = kparams.get('_in_variable_names', {})

                    arg_count = f.func_code.co_argcount
                    param_names = f.func_code.co_varnames[1:arg_count]

                    try:
                        in_params = TypeInfo()

                        for i in range(len(params)):
                            e0 = _in_variable_names.get(param_names[i],
                                                                 param_names[i])
                            e1 = params[i]

                            in_params[e0] = e1

                    except IndexError, e:
                        raise Exception("%s has parameter numbers mismatching" %
                                                                    f.func_name)

                    message=Message.produce(type_name=_in_message, namespace=ns,
                                                            members=in_params)
                    message.__namespace__ = ns
                    message.resolve_namespace(message, ns)
                    return message

                def get_output_message(ns):
                    _returns = kparams.get('_returns')

                    _out_message = kparams.get('_out_message', '%sResponse' %
                                                                    f.func_name)

                    kparams.get('_out_variable_name')
                    out_params = TypeInfo()

                    if _returns:
                        if isinstance(_returns, (list, tuple)):
                            default_names = ['%sResult%d' % (f.func_name, i)
                                                  for i in range(len(_returns))]

                            _out_variable_names = kparams.get(
                                        '_out_variable_names', default_names)

                            assert (len(_returns) == len(_out_variable_names))

                            var_pair = zip(_out_variable_names,_returns)
                            out_params = TypeInfo(var_pair)

                        else:
                            _out_variable_name = kparams.get(
                                 '_out_variable_name', '%sResult' % f.func_name)

                            out_params[_out_variable_name] = _returns

                    message=Message.produce(type_name=_out_message,namespace=ns,
                                                             members=out_params)
                    message.__namespace__ = ns
                    message.resolve_namespace(message, ns)
                    return message

                _is_callback = kparams.get('_is_callback', False)
                _public_name = kparams.get('_public_name', f.func_name)
                _is_async = kparams.get('_is_async', False)
                if _is_async:
                    logger.warning("Async methods are not supported in this "
                                   "release. _is_async is ignored.")
                    _is_async = False

                _mtom = kparams.get('_mtom', False)
                _in_header = kparams.get('_in_header', None)
                _out_header = kparams.get('_out_header', None)

                # the decorator function does not have a reference to the
                # class and needs to be passed in
                ns = kwargs['clazz'].get_tns()

                in_message = get_input_message(ns)
                out_message = get_output_message(ns)

                if not (_in_header is None):
                    _in_header.resolve_namespace(_in_header, ns)
                if not (_out_header is None):
                    _out_header.resolve_namespace(_out_header, ns)

                doc = getattr(f, '__doc__')
                descriptor = MethodDescriptor(f.func_name, _public_name,
                        in_message, out_message, doc, _is_callback, _is_async,
                        _mtom, _in_header, _out_header)

                return descriptor

            return f(*args, **kwargs)

        explain_method.__doc__ = f.__doc__
        explain_method._is_rpc = True
        explain_method.func_name = f.func_name

        return explain_method

    return explain

class _SchemaInfo(object):
    def __init__(self):
        self.elements = odict()
        self.types = odict()

class _SchemaEntries(object):
    def __init__(self, tns):
        self.namespaces = odict()
        self.imports = {}
        self.tns = tns

    def has_class(self, cls):
        retval = False
        ns_prefix = cls.get_namespace_prefix()

        if ns_prefix in cocy.soaplib.const_nsmap:
            retval = True

        else:
            type_name = cls.get_type_name()

            if (ns_prefix in self.namespaces) and \
                              (type_name in self.namespaces[ns_prefix].types):
                retval = True

        return retval

    def get_schema_info(self, prefix):
        if prefix in self.namespaces:
            schema = self.namespaces[prefix]
        else:
            schema = self.namespaces[prefix] = _SchemaInfo()

        return schema

    # FIXME: this is an ugly hack. we need proper dependency management
    def __check_imports(self, cls, node):
        pref_tns = cls.get_namespace_prefix()

        def is_valid_import(pref):
            return not (
                (pref in cocy.soaplib.const_nsmap) or (pref == pref_tns)
            )

        if not (pref_tns in self.imports):
            self.imports[pref_tns] = set()

        for c in node:
            if c.tag == "{%s}complexContent" % cocy.soaplib.ns_xsd:
                seq = c.getchildren()[0].getchildren()[0] # FIXME: ugly, isn't it?

                extension = c.getchildren()[0]
                if extension.tag == '{%s}extension' % cocy.soaplib.ns_xsd:
                    pref = extension.attrib['base'].split(':')[0]
                    if is_valid_import(pref):
                        self.imports[pref_tns].add(soaplib.nsmap[pref])
            else:
                seq = c

            if seq.tag == '{%s}sequence' % cocy.soaplib.ns_xsd:
                for e in seq:
                    pref = e.attrib['type'].split(':')[0]
                    if is_valid_import(pref):
                        self.imports[pref_tns].add(soaplib.nsmap[pref])

            elif seq.tag == '{%s}restriction' % cocy.soaplib.ns_xsd:
                pref = seq.attrib['base'].split(':')[0]
                if is_valid_import(pref):
                    self.imports[pref_tns].add(soaplib.nsmap[pref])

            else:
                raise Exception("i guess you need to hack some more")

    def add_element(self, cls, node):
        schema_info = self.get_schema_info(cls.get_namespace_prefix())
        schema_info.elements[cls.get_type_name()] = node

    def add_simple_type(self, cls, node):
        self.__check_imports(cls, node)
        schema_info = self.get_schema_info(cls.get_namespace_prefix())
        schema_info.types[cls.get_type_name()] = node

    def add_complex_type(self, cls, node):
        self.__check_imports(cls, node)
        schema_info = self.get_schema_info(cls.get_namespace_prefix())
        schema_info.types[cls.get_type_name()] = node


_public_methods_cache = {}

class DefinitionBase(object):
    '''
    This class serves as the base for all soap services.  Subclasses of this
    class will use the rpc decorator to flag methods to be exposed via soap.
    This class is responsible for generating the wsdl for this service
    definition.

    It is a natural abstract base class, because it's of no use without any
    method definitions, hence the 'Base' suffix in the name.
    '''

    __tns__ = None
    __in_header__ = None
    __out_header__ = None

    def __init__(self, environ=None):
        self.soap_in_header = None
        self.soap_out_header = None

        cls = self.__class__
        if not (cls in _public_methods_cache):
            _public_methods_cache[cls] = self.build_public_methods()

        self.public_methods = _public_methods_cache[cls]

    def on_method_call(self, environ, method_name, py_params, soap_params):
        '''
        Called BEFORE the service implementing the functionality is called
        @param the wsgi environment
        @param the method name
        @param the body element of the soap request
        @param the tuple of python params being passed to the method
        @param the soap elements for each params
        '''
        pass

    def on_method_return(self, environ, py_results, soap_results,
                                                             http_resp_headers):
        '''
        Called AFTER the service implementing the functionality is called
        @param the wsgi environment
        @param the python results from the method
        @param the xml element containing the return value(s) from the method
        @param http response headers as a dict of strings
        '''
        pass

    def on_method_exception_object(self, environ, exc):
        '''
        Called BEFORE the exception is serialized, when an error occurs durring
        execution
        @param the wsgi environment
        @param the exception object
        '''
        pass

    def on_method_exception_xml(self, environ, fault_xml):
        '''
        Called AFTER the exception is serialized, when an error occurs durring
        execution
        @param the wsgi environment
        @param the xml element containing the exception object serialized to a
        soap fault
        '''
        pass

    def call_wrapper(self, call, params):
        '''
        Called in place of the original method call.
        @param the original method call
        @param the arguments to the call
        '''
        return call(*params)

    @classmethod
    def get_tns(cls):
        if not (cls.__tns__ is None):
            return cls.__tns__

        service_name = cls.__name__.split('.')[-1]

        retval = '.'.join((cls.__module__, service_name))
        if cls.__module__ == '__main__':
            retval = '.'.join((service_name, service_name))

        return retval

    def build_public_methods(self):
        '''Returns a list of method descriptors for this object'''

        logger.debug('building public methods')
        public_methods = []

        for func_name in dir(self):
            if func_name == 'public_methods':
                continue
            func = getattr(self, func_name)
            if callable(func) and hasattr(func, '_is_rpc'):
                descriptor = func(_method_descriptor=True, clazz=self.__class__)
                public_methods.append(descriptor)

        return public_methods

    def get_method(self, name):
        '''
        Returns the metod descriptor based on element name or soap action
        '''

        for method in self.public_methods:
            type_name = method.in_message.get_type_name()
            if '{%s}%s' % (self.get_tns(), type_name) == name:
                return method

        for method in self.public_methods:
            if method.public_name == name:
                return method

        raise Exception('Method "%s" not found' % name)

    def get_service_names(self):
        '''
        Returns the service name(s) for this service. If this
        object has callbacks, then a second service is declared in
        the wsdl for those callbacks
        '''

        service_name = self.__class__.__name__.split('.')[-1]

        if self._hasCallbacks():
            return [service_name, '%sCallback' % service_name]

        return [service_name]

    def add_port_type(self, root, service_name, types, url, port_type):
        ns_wsdl = cocy.soaplib.ns_wsdl

        for method in self.public_methods:
            operation = etree.SubElement(port_type,'{%s}operation' % ns_wsdl)
            operation.set('name', method.name)

            if method.doc is not None:
                documentation = etree.SubElement(operation, '{%s}documentation'
                                                                % ns_wsdl)
                documentation.text = method.doc

            operation.set('parameterOrder', method.in_message.get_type_name())

            op_input = etree.SubElement(operation, '{%s}input' % ns_wsdl)
            op_input.set('name', method.in_message.get_type_name())
            op_input.set('message', method.in_message.get_type_name_ns())

            op_output = etree.SubElement(operation, '{%s}output' %  ns_wsdl)
            op_output.set('name', method.out_message.get_type_name())
            op_output.set('message', method.out_message.get_type_name_ns())

    def add_schema(self, schema_entries):
        '''
        A method for adding the appropriate entries
        to the schema for the types in the specified methods.

        @param the schema node to add the schema elements to. if it is None,
               the schema nodes are returned inside a dictionary
        @param the schema node dictinary, where keys are prefixes of the schema
               stored schema node
        '''

        if schema_entries is None:
            schema_entries = _SchemaEntries(self.get_tns())

        if self.__in_header__ != None:
            self.__in_header__.resolve_namespace(self.__in_header__, self.get_tns())
            self.__in_header__.add_to_schema(schema_entries)

        if self.__out_header__ != None:
            self.__out_header__.resolve_namespace(self.__out_header__, self.get_tns())
            self.__out_header__.add_to_schema(schema_entries)

        for method in self.public_methods:
            method.in_message.add_to_schema(schema_entries)
            method.out_message.add_to_schema(schema_entries)

            if method.in_header is None:
                method.in_header = self.__in_header__
            else:
                method.in_header.add_to_schema(schema_entries)

            if method.out_header is None:
                method.out_header = self.__out_header__
            else:
                method.out_header.add_to_schema(schema_entries)

        return schema_entries

    def __add_message_for_object(self, root, messages, obj):
        if obj != None and not (obj.get_type_name() in messages):
            messages.add(obj.get_type_name())

            message = etree.SubElement(root, '{%s}message' % cocy.soaplib.ns_wsdl)
            message.set('name', obj.get_type_name())

            part = etree.SubElement(message, '{%s}part' % cocy.soaplib.ns_wsdl)
            part.set('name', obj.get_type_name())
            part.set('element', obj.get_type_name_ns())

    def add_messages_for_methods(self, root, messages):
        '''
        A private method for adding message elements to the wsdl
        @param the the root element of the wsdl
        '''

        for method in self.public_methods:
            self.__add_message_for_object(root, messages, method.in_message)
            self.__add_message_for_object(root, messages, method.out_message)
            self.__add_message_for_object(root, messages, method.in_header)
            self.__add_message_for_object(root, messages, method.out_header)

    def add_bindings_for_methods(self, root, service_name, types, url, binding):
        '''
        A private method for adding bindings to the wsdl

        @param the root element of the wsdl
        @param the name of this service
        '''

        ns_wsdl = cocy.soaplib.ns_wsdl
        ns_soap = cocy.soaplib.ns_soap

        for method in self.public_methods:
            operation = etree.SubElement(binding, '{%s}operation' % ns_wsdl)
            operation.set('name', method.name)

            soap_operation = etree.SubElement(operation, '{%s}operation' %
                                                                       ns_soap)
            soap_operation.set('soapAction', method.public_name)
            soap_operation.set('style', 'document')

            # get input
            input = etree.SubElement(operation, '{%s}input' % ns_wsdl)
            input.set('name', method.in_message.get_type_name())

            soap_body = etree.SubElement(input, '{%s}body' % ns_soap)
            soap_body.set('use', 'literal')

            # get input soap header
            in_header = method.in_header
            if in_header is None:
                in_header = self.__in_header__

            if not (in_header is None):
                soap_header = etree.SubElement(input, '{%s}header' % ns_soap)
                soap_header.set('use', 'literal')
                soap_header.set('message', in_header.get_type_name_ns())
                soap_header.set('part', in_header.get_type_name())

            output = etree.SubElement(operation, '{%s}output' % ns_wsdl)
            output.set('name', method.out_message.get_type_name())

            soap_body = etree.SubElement(output, '{%s}body' % ns_soap)
            soap_body.set('use', 'literal')

            # get input soap header
            out_header = method.in_header
            if out_header is None:
                out_header = self.__in_header__

            if not (out_header is None):
                soap_header = etree.SubElement(output, '{%s}header' % ns_soap)
                soap_header.set('use', 'literal')
                soap_header.set('message', out_header.get_type_name_ns())
                soap_header.set('part', out_header.get_type_name())



"""
..
   This file is part of the CoCy program.
   Copyright (C) 2011 Michael N. Lipp
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

.. codeauthor:: mnl
"""
from cocy.providers import Manifest, MediaPlayer
from cocy import providers
from circuits.core.handlers import handler
from circuits_bricks.core.timers import Timer
import time

class DummyPlayer(providers.MediaPlayer):
    '''
    classdocs
    '''

    manifest = Manifest("Dummy Player", "CoCy Dummy Media Player")

    def __init__(self):
        super(DummyPlayer, self).__init__(self.manifest)
        self._timer = None

    @handler("provider_updated")
    def _on_provider_updated_handler(self, provider, changed):
        if "state" in changed:
            state = changed["state"]
            if state == "PLAYING":
                self.current_track_duration = 60
                if self._timer:
                    self._timer.unregister()
                self._timer = Timer(self.current_track_duration, 
                                    MediaPlayer.EndOfMedia()).register(self)
            elif state == "IDLE":
                if self._timer:
                    self._timer.unregister()

    def current_position(self):
        if self._timer is None:
            return None
        return self._current_track_duration - (self._timer.expiry - time.time())



"""
..
   This file is part of the CoCy program.
   Copyright (C) 2012 Michael N. Lipp
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

.. codeauthor:: mnl
"""
from circuits_minpor import Portlet
from tenjin.helpers import *
from circuits_minpor.portlet import TemplatePortlet
import os
from circuits.core.handlers import handler
from circuits_minpor.portal import PortalUpdate

class BinaryLightPortlet(TemplatePortlet):

    def __init__(self, binary_light):
        super(BinaryLightPortlet, self) \
            .__init__(os.path.dirname(__file__), "binary_light", weight=1)
        self._binary_light = binary_light
        @handler("provider_updated", channel=binary_light.channel)
        def _on_provider_updated(self, provider, changed):
            self._on_updated(provider, changed)
        self.addHandler(_on_provider_updated)

    def description(self, locales=[]):
        return Portlet.Description\
            (self._handle, self.translation(locales) \
             .ugettext("Binary Light"))

    def do_render(self, mime_type, mode, window_state, locales, url_generator, 
                  invocation_id, context_exts = {}, globs_exts = {}, **kwargs):
        return super(BinaryLightPortlet, self)\
            .do_render(mime_type, mode, window_state, locales, url_generator,
                       invocation_id, context_exts =
                       { "binary_light": self._binary_light })

    @handler("portlet_added")
    def _on_portlet_added(self, portal, portlet):
        self._portal_channel = portal.channel
        @handler("portal_client_connect", channel=portal.channel)
        def _on_client_connect(*args):
            self._on_updated(self._binary_light, None)
        self.addHandler(_on_client_connect)

    def _on_updated(self, provider, changed):
        if provider != self._binary_light:
            return
        self.fire(PortalUpdate(self, "new_state", self._binary_light.state),
                   self._portal_channel)



"""
..
   This file is part of the CoCy program.
   Copyright (C) 2011 Michael N. Lipp
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

.. codeauthor:: mnl
"""
from circuits.core.components import BaseComponent
from circuits.core.events import Event
from circuits.core.handlers import handler

from soaplib.soap import from_soap

class SOAPOperation(Event):
    """Soap Operation Event"""

class SOAP(BaseComponent):
    '''
    classdocs
    '''

    channel = "web"

    def __init__(self, path=None, channel="*", encoding="utf-8"):
        super(SOAP, self).__init__()

        self.path = path
        self.channel = channel
        self.encoding = encoding

    @handler("request", filter=True, priority=0.1)
    def _on_request(self, request, response):
        if self.path is not None and self.path != request.path.rstrip("/"):
            return

        try:
            # Test if this is a SOAP request. SOAP 1.1 specifies special
            # header, SOAP 1.2 special Content-Type
            soapAction = request.headers["SOAPAction"];
            import cgi
            contentType = cgi.parse_header(request.headers["Content-Type"]);
            if (not soapAction and contentType[0] != "application/soap+xml"):
                return
            # Get body data of request
            body = request.body.read()
            # Use soaplib to separate header and payload
            charset = contentType[1].get('charset',None)
            if charset is None:
                charset = 'ascii'
            
            payload, soapheader = from_soap(body, charset)
            from soaplib.soap import collapse_swa
            payload = collapse_swa(contentType, payload)

            if payload is not None:
                soapAction = payload.tag

            for node in payload:
                print node

            response.headers["Content-Type"] = "text/xml"

            #value = self.push(SoapOperation(*params), c, t)
            #value.response = response
            #value.onSet = ("value_changed", self)
        except Exception as e:
            # TODO: 
            r = self._error(1, "%s: %s" % (type(e), e))
            return r
        else:
            return True



"""
..
   This file is part of the CoCy program.
   Copyright (C) 2011 Michael N. Lipp
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from circuits.core.components import BaseComponent
from circuits_bricks.misc.compquery import Queryable
from uuid import uuid4
from xml.etree.ElementTree import ElementTree, Element, SubElement, QName
from circuits_bricks.web.dispatchers.dispatcher import ScopedChannel
from cocy.upnp import SSDP_DEVICE_SCHEMA, SSDP_SCHEMAS, UPNP_EVENT_NS,\
    UPNP_SERVICE_ID_PREFIX, SERVER_HELLO
from circuits.web.controllers import Controller, expose, BaseController
from cocy.misc import parseSoapRequest, splitQTag, buildSoapResponse
from cocy.upnp.device_server import UPnPError
from email.utils import formatdate
from circuits_bricks.core.timers import Timer
from circuits.core.events import Event
from circuits.core.handlers import handler
from inspect import getmembers, ismethod
from StringIO import StringIO
from circuits_bricks.web.client import Client, Request
from cocy.upnp.service import UPnPService
from circuits_bricks.app.logger import Log
import logging
from cocy import misc

class UPnPServiceError(Exception):
    
    def __init__(self, code):
        self._code = code
        
    @property
    def code(self):
        return self._code
    

class UPnPDeviceAdapter(BaseComponent, Queryable):
    """
    This class publishes a :class:`cocy.Provider` as a UPnP device.
    
    The class itself holds and provides the basic information that is 
    required to announce the the device using SSDP. The detailed description
    is provided using a child :class:`~.UPnPDeviceController`.
    Actions can be invoked on the provider one or more child
    components of type :class:`~.UPnPServiceController`. 
    """

    channel = "upnp"
    
    class DeviceProperties(object):
        """
        This class simply serves as a container for common information about 
        UPnP device types.
        """
        
        def __init__(self, dev_type, ver, spec_ver_major, spec_ver_minor,
                     services, desc_gen):
            object.__init__(self)
            self.type = dev_type
            self.ver = ver
            self.spec_ver_major = spec_ver_major
            self.spec_ver_minor = spec_ver_minor
            self.services = services
            self.desc_gen = desc_gen

    _service_registry = dict()
    _mapping = dict()
    _props = DeviceProperties("Undefined", 0, 0, 0, [], None)
    """
    This dictionary maps the defined :class:`cocy.providers.Provider`
    classes to UPnP device properties.
    """

    @classmethod
    def add_adapter(cls, provider_class, props):
        # Create and register all known service types
        cls._mapping[provider_class] = props

    def __init__(self, server, provider, config_id, uuid_map, port):
        super(UPnPDeviceAdapter, self).__init__();
        
        # It may turn out that no adapter can be constructed
        self.valid = False
        # Try to find a match UPnP info for the device 
        for itf, props in self._mapping.iteritems():
            if isinstance(provider, itf):
                self.valid = True
                self._props = props
                break;
        if not self.valid:
            return
        # Remember the provider as our "model"
        self._provider = provider
        # Remember port
        self._web_server_port = port
        # Get instance information about the provider
        manifest = provider.provider_manifest
        if manifest.unique_id and uuid_map.has_key(manifest.unique_id):
            self._uuid = uuid_map[manifest.unique_id]
        else:
            self._uuid = str(uuid4())
            uuid_map[manifest.unique_id] = self._uuid
        # Generate a unique path that will be used to access this device
        self._path = "/" + self.uuid
        # Remember the configuration id
        self.config_id = config_id

        # Assemble the services supported by the provider
        self._services = set()
        service_insts = []
        for (service_type, service_id, controller) in self._props.services:
            if not UPnPDeviceAdapter._service_registry.has_key(service_type):
                try:
                    service = UPnPService \
                        (config_id, service_type).register(self) \
                        .register(server)
                except:
                    continue
            else:
                service = UPnPDeviceAdapter._service_registry[service_type]
            self._services.add(service)
            service_insts.append((service, service_id))
            # Create an adapter that links this class's service with the web
            # component interface of circuits
            controller(self, self._path, service, service_id).register(self)

        # Create an adapter that links this class with the web
        # component interface of circuits
        UPnPDeviceController(ScopedChannel("upnp-web", self._path),
                             self, config_id, props, service_insts) \
            .register(self)

    @property
    def provider(self):
        return getattr(self, "_provider", None)

    @property
    def path(self):
        return getattr(self, "_path", None)

    @property
    def web_server_port(self):
        return getattr(self, "_web_server_port", None)

    @property
    def uuid(self):
        return getattr(self, "_uuid", None)

    @property
    def root_device(self):
        return True # TODO:

    @property
    def services(self):
        return getattr(self, "_services", None)

    def __getattr__(self, name):
        if not name.startswith("_") and hasattr(self._props, name):
            return getattr(self._props, name, None)
        raise AttributeError("No " + name + " in " + self.__class__.__name__)

    @property
    def type_ver(self):
        return "%s:%s" % (str(self.type), str(self.ver))


class UPnPDeviceController(Controller):

    def __init__(self, channel, adapter, config_id, props, service_insts):
        super(UPnPDeviceController, self).__init__(channel=channel);
        # Generate a device description for the device
        desc = getattr(self, props.desc_gen)\
            (adapter, config_id, props, service_insts)
        misc.set_ns_prefixes(desc, { "": SSDP_DEVICE_SCHEMA })
        writer = StringIO()
        writer.write("<?xml version='1.0' encoding='utf-8'?>")
        ElementTree(desc).write(writer, encoding="utf-8")
        self.description = writer.getvalue()

    def _common_device_desc(self, adapter, config_id, props, services):
        root = Element("{%s}root" % SSDP_DEVICE_SCHEMA,
                       attrib = {"configId": str(config_id)})
        specVersion = SubElement(root, "{%s}specVersion" % SSDP_DEVICE_SCHEMA)
        SubElement(specVersion, "{%s}major" % SSDP_DEVICE_SCHEMA).text \
            = str(props.spec_ver_major)
        SubElement(specVersion, "{%s}minor" % SSDP_DEVICE_SCHEMA).text \
            = str(props.spec_ver_minor)
        device = SubElement(root, "{%s}device" % SSDP_DEVICE_SCHEMA)
        SubElement(device, "{%s}deviceType" % SSDP_DEVICE_SCHEMA).text \
            = SSDP_SCHEMAS + ":device:" + adapter.type_ver
        manifest = adapter.provider.provider_manifest
        for tag, value in \
            [("friendlyName", manifest.display_name),
             ("manufacturer", manifest.manufacturer or "cocy"),
             ("modelDescription", manifest.description),
             ("modelName", manifest.full_name or manifest.display_name),
             ("modelNumber", manifest.model_number)]:
            if value is not None:
                SubElement(device, "{%s}%s" % (SSDP_DEVICE_SCHEMA, tag)).text \
                    = value
        SubElement(device, "{%s}UDN" % SSDP_DEVICE_SCHEMA).text \
            = "uuid:" + adapter.uuid
        if services and len(services) > 0:
            serviceList = SubElement(device, 
                                     "{%s}serviceList" % SSDP_DEVICE_SCHEMA)
            for service in services:
                self._describeService(adapter, serviceList, service)
        return root
        
    def _describeService(self, adapter, service_list, service_tuple):
        (service_type, service_id) = service_tuple
        service = SubElement(service_list, "{%s}service" % SSDP_DEVICE_SCHEMA)
        SubElement(service, "{%s}serviceType" % SSDP_DEVICE_SCHEMA).text \
            = SSDP_SCHEMAS + ":service:" + service_type.type_ver
        SubElement(service, "{%s}serviceId" % SSDP_DEVICE_SCHEMA).text \
            = UPNP_SERVICE_ID_PREFIX + service_id
        SubElement(service, "{%s}SCPDURL" % SSDP_DEVICE_SCHEMA).text \
            = service_type.description_url
        SubElement(service, "{%s}controlURL" % SSDP_DEVICE_SCHEMA).text \
            = adapter.path + "/" + service_id + "/control"
        SubElement(service, "{%s}eventSubURL" % SSDP_DEVICE_SCHEMA).text \
            = adapter.path + "/" + service_id + "/sub"

    @expose("description.xml")
    def description(self, *args):
        self.response.headers["Content-Type"] = "text/xml"
        return self.description


class Notification(Event):
    pass


class UPnPSubscription(BaseController):
    
    def __init__(self, callbacks, timeout, protocol):
        self._uuid = str(uuid4())
        super(UPnPSubscription, self).__init__(channel="subs:" + self._uuid)
        self._callbacks = callbacks
        self._used_callback = 0
        self._client = Client(self._callbacks[self._used_callback], 
                              self.channel).register(self)
        self._protocol = protocol
        self._seq = 0
        if timeout > 0:
            self._expiry_timer = Timer \
                (timeout, Event.create("upnp_subs_end"), self).register(self)

    @handler("registered")
    def _on_registered(self, component, parent):
        if component != self:
            return
        @handler("notification", channel=parent.notification_channel)
        def _on_notification_handler(self, state_vars):
            self._on_notification(state_vars)
        self.addHandler(_on_notification_handler)
        state_vars = dict()
        for name, method in getmembers \
            (self.parent, lambda x: ismethod(x) and hasattr(x, "_evented_by")):
            state_vars[name] = method()
        if len(state_vars) > 0:
            self._on_notification(state_vars)
        self.fire(Log(logging.DEBUG, "Subscribtion for " + str(self._callbacks)
                      + " on " + self.parent.notification_channel 
                      + " created"), "logger")

    def _on_notification(self, state_vars):
        root = Element(QName(UPNP_EVENT_NS, "propertyset"))
        for name, value in state_vars.items():
            prop = SubElement(root, QName(UPNP_EVENT_NS, "property"))
            val = SubElement(prop, QName(UPNP_EVENT_NS, name))
            if isinstance(value, bool):
                val.text = "1" if value else "0"
            else:
                val.text = str(value)
        misc.set_ns_prefixes(root, { "": UPNP_EVENT_NS })
        writer = StringIO()
        writer.write("<?xml version='1.0' encoding='utf-8'?>")
        ElementTree(root).write(writer, encoding="utf-8")
        body = writer.getvalue()
        self.fire(Log(logging.DEBUG, "Notifying " 
                      + self._callbacks[self._used_callback]
                      + " about " + str(state_vars)), "logger")
        self.fire(Request("NOTIFY", self._callbacks[self._used_callback], body,
                          { "CONTENT-TYPE": "text/xml; charset=\"utf-8\"",
                            "NT": "upnp:event",
                            "NTS": "upnp:propchange",
                            "SID": self.sid,
                            "SEQ": self._seq }))
        self._seq += 1

    @handler("upnp_subs_end")
    def _on_subs_end(self):
        self.unregister()
        self.fire(Log(logging.DEBUG, "Subscribtion for " + str(self._callbacks)
                      + " on " + self.parent.notification_channel
                      + " cancelled"), "logger")

    @handler("upnp_subs_renewal")
    def _on_renewal(self, timeout):
        self._expiry_timer.interval = timeout
        self._expiry_timer.reset()
        self.fire(Log(logging.DEBUG, "Subscribtion for " + str(self._callbacks)
                      + " on " + self.parent.notification_channel
                      + " renewed"), "logger")

    @property
    def sid(self):
        return "uuid:" + self._uuid

    @classmethod
    def sid2chan(cls, sid):
        return "subs:" + sid[5:]


class UPnPServiceController(BaseController):

    def __init__ \
        (self, adapter, device_path, service, service_id):
        super(UPnPServiceController, self).__init__ \
            (channel=ScopedChannel("upnp-web", device_path + "/" + service_id));
        self._service = service
        self._notification_channel = adapter.uuid + "/" \
            + service_id + "/notifications"

    @property
    def notification_channel(self):
        return getattr(self, "_notification_channel", None)

    @handler("registered")
    def _on_registered(self, component, parent):
        if component != self:
            return
        @handler("provider_updated", channel=self.parent.provider.channel)
        def _on_provider_updated_handler(self, provider, changed):
            if provider != self.parent.provider:
                return
            self._on_provider_updated(changed)
        self.addHandler(_on_provider_updated_handler)
            
    def _on_provider_updated(self, changed):
        state_vars = dict()
        for name, method in getmembers \
            (self, lambda x: ismethod(x) and hasattr(x, "_evented_by")):
            if method._evented_by is not None:
                state_vars[name] = changed[method._evented_by]
        if len(state_vars) > 0:
            self.fire(Notification(state_vars), self.notification_channel)

    @expose("control")
    def _control(self, *args):
        payload = parseSoapRequest(self.request)[2]
        action_ns, action = splitQTag(payload.tag)
        action_args = dict()
        for node in payload:
            action_args[node.tag] = node.text
        method = getattr(self, action, None)
        if method is None or not getattr(method, "_is_upnp_service", False):
            self.fire(Log(logging.INFO, 'Action ' + action 
                          + " not implemented"), "logger")
            return UPnPError(self.request, self.response, 401)
        try:
            out_args = method(**action_args)
        except UPnPServiceError as error:
            return UPnPError(self.request, self.response, error.code)
        result = Element("{%s}%sResponse" % (action_ns, action))
        for name, value in out_args:
            arg = SubElement(result, name)
            arg.text = str(value)
        return buildSoapResponse(self.response, result)

    @expose("sub")
    def _sub(self, *args):
        if self.request.method == "SUBSCRIBE":
            timeout = self.request.headers["Timeout"]
            timeout = timeout[len("Second-"):]
            try:
                timeout = int(timeout)
            except ValueError:
                timeout = 1800
            self.response.headers["Date"] = formatdate(usegmt=True)
            self.response.headers["Server"] = SERVER_HELLO
            self.response.headers["Timeout"] = "Second-" + str(timeout)
            if "SID" in self.request.headers:
                # renewal
                sid = self.request.headers["SID"]
                self.response.headers["SID"] = sid
                self.fire(Event.create("upnp_subs_renewal", timeout), 
                          UPnPSubscription.sid2chan(sid))
                return ""
            callbacks = []
            for cb in self.request.headers["CALLBACK"].split("<")[1:]:
                callbacks.append(cb[:cb.rindex(">")])
            subs = UPnPSubscription(callbacks, timeout, 
                                    self.request.protocol).register(self)
            self.response.headers["SID"] = subs.sid
            return ""
        elif self.request.method == "UNSUBSCRIBE":
            sid = self.request.headers["SID"]
            self.fire(Event.create("upnp_subs_end"), 
                      UPnPSubscription.sid2chan(sid))
            return ""


def upnp_service(f):
    setattr(f, "_is_upnp_service", True)
    return f

def upnp_state(*args, **kwargs):
    if len(args) == 0 or not hasattr(args[0], "__call__"):
        # function to be wrapped isn't first parameter, re-wrap
        def wrapper(f):
            setattr(f, "_is_upnp_state", True)
            if "evented_by" in kwargs:
                setattr(f, "_evented_by", kwargs["evented_by"])
            return f
        return wrapper
    else:
        # function to be wrapped is first parameter
        f = args[0]
        setattr(f, "_is_upnp_state", True)
        return f



"""
..
   This file is part of the CoCy program.
   Copyright (C) 2011 Michael N. Lipp
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import platform

SSDP_PORT = 1900
SSDP_ADDR = '239.255.255.250'

SSDP_SCHEMAS = "urn:schemas-upnp-org"
SSDP_DEVICE_SCHEMA = "urn:schemas-upnp-org:device-1-0"
UPNP_SERVICE_SCHEMA = "urn:schemas-upnp-org:service-1-0"
UPNP_ROOTDEVICE = "upnp:rootdevice"
UPNP_SERVICE_ID_PREFIX = "urn:upnp-org:serviceId:"
UPNP_CONTROL_NS = "urn:schemas-upnp-org:control-1-0"
UPNP_EVENT_NS = "urn:schemas-upnp-org:event-1-0"
UPNP_AVT_EVENT_NS = "urn:schemas-upnp-org:metadata-1-0/AVT/"
UPNP_RCS_EVENT_NS = "urn:schemas-upnp-org:metadata-1-0/RCS/"
COCY_SERVICE_EXT = "urn:cocy-service-ext"

SERVER_HELLO = (platform.system() + '/' + platform.release()
                + " UPnP/1.1 CoCy/0.1")

from .device_server import UPnPDeviceServer
from .adapters.adapter import UPnPDeviceAdapter
from .ssdp import SSDPTranceiver

__all__ = ["SSDP_PORT", "SSDP_ADDR", "SSDP_SCHEMAS",
           "SSDP_DEVICE_SCHEMA", "UPNP_SERVICE_ID_PREFIX", 
           "UPnPDeviceServer", "SSDPTranceiver"
]




#
# soaplib - Copyright (C) cocy.soaplib contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#
# Adapted to standard etree implementation (mnl at mnl.de)

import base64
import cStringIO

from cocy.soaplib.serializers.base import Base
from cocy.soaplib.serializers import nillable_value, nillable_element

from xml import etree

class Attachment(Base):
    __type_name__ = 'base64Binary'
    __namespace__ = "http://www.w3.org/2001/XMLSchema"

    def __init__(self, data=None, file_name=None):
        self.data = data
        self.file_name = file_name

    def save_to_file(self):
        '''
        This method writes the data to the specified file.  This method
        assumes that the file_name is the full path to the file to be written.
        This method also assumes that self.data is the base64 decoded data,
        and will do no additional transformations on it, simply write it to
        disk.
        '''

        if not self.data:
            raise Exception("No data to write")

        if not self.file_name:
            raise Exception("No file_name specified")

        f = open(self.file_name, 'wb')
        f.write(self.data)
        f.close()

    def load_from_file(self):
        '''
        This method loads the data from the specified file, and does
        no encoding/decoding of the data
        '''
        if not self.file_name:
            raise Exception("No file_name specified")
        f = open(self.file_name, 'rb')
        self.data = f.read()
        f.close()

    @classmethod
    @nillable_value
    def to_xml(cls, value, tns, parent_elt, name='retval'):
        '''
        This class method takes the data from the attachment and
        base64 encodes it as the text of an Element. An attachment can
        specify a file_name and if no data is given, it will read the data
        from the file
        '''

        assert isinstance(value, cls)

        element = etree.ElementTree.SubElement(parent_elt, '{%s}%s' % (tns,name))
        if value.data:
            # the data has already been loaded, just encode
            # and return the element
            element.text = base64.encodestring(value.data)

        elif value.file_name:
            # the data hasn't been loaded, but a file has been
            # specified
            data_string = cStringIO.StringIO()

            file_name = value.file_name
            file = open(file_name, 'rb')
            base64.encode(file, data_string)
            file.close()

            # go back to the begining of the data
            data_string.seek(0)
            element.text = str(data_string.read())

        else:
            raise Exception("Neither data nor a file_name has been specified")

    @classmethod
    @nillable_element
    def from_xml(cls, element):
        '''
        This method returns an Attachment object that contains
        the base64 decoded string of the text of the given element
        '''
        data = base64.decodestring(element.text)
        a = Attachment(data=data)
        return a




#
# soaplib - Copyright (C) cocy.soaplib contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#
# Adapted to standard etree implementation (mnl at mnl.de)

import logging
logger = logging.getLogger(__name__)

from xml import etree

from base64 import b64encode
from urllib import unquote

# import email data format related stuff
try:
    # python >= 2.5
    from email.mime.multipart import MIMEMultipart
    from email.mime.application import MIMEApplication
    from email.encoders import encode_7or8bit
except ImportError:
    # python 2.4
    from email.MIMENonMultipart import MIMENonMultipart
    from email.MIMEMultipart import MIMEMultipart
    from email.Encoders import encode_7or8bit

from email import message_from_string

# import cocy.soaplib stuff
from cocy.soaplib.serializers.binary import Attachment
from cocy.soaplib.serializers.clazz import ClassSerializer

import cocy.soaplib

class Message(ClassSerializer):
    pass


from cocy.soaplib.serializers.exception import Fault

class MethodDescriptor(object):
    '''
    This class represents the method signature of a soap method,
    and is returned by the soapdocument, or rpc decorators.
    '''

    def __init__(self, name, public_name, in_message, out_message, doc,
                 is_callback=False, is_async=False, mtom=False, in_header=None,
                 out_header=None):

        self.name = name
        self.public_name = public_name
        self.in_message = in_message
        self.out_message = out_message
        self.doc = doc
        self.is_callback = is_callback
        self.is_async = is_async
        self.mtom = mtom
        self.in_header = in_header
        self.out_header = out_header

def from_soap(xml_string, http_charset):
    '''
    Parses the xml string into the header and payload
    '''
    try:
        root, xmlids = etree.ElementTree.XMLID(xml_string.decode(http_charset))
    except ValueError,e:
        logger.debug('%s -- falling back to str decoding.' % (e))
        root, xmlids = etree.ElementTree.XMLID(xml_string)

    if xmlids:
        resolve_hrefs(root, xmlids)

    if root.tag != '{%s}Envelope' % cocy.soaplib.ns_soap_env:
        raise Fault('Client.SoapError', 'No {%s}Envelope element was found!' % cocy.soaplib.ns_soap_env)

    header_envelope = root.findall('{%s}Header' % cocy.soaplib.ns_soap_env)
    body_envelope = root.findall('{%s}Body' % cocy.soaplib.ns_soap_env)

    if len(header_envelope) == 0 and len(body_envelope) == 0:
        raise Fault('Client.SoapError', 'Soap envelope is empty!' % cocy.soaplib.ns_soap_env)
    body=None
    if len(body_envelope) > 0 and len(body_envelope[0]) > 0:
        body = body_envelope[0].getchildren()[0]

    header=None
    if len(header_envelope) > 0 and len(header_envelope[0]) > 0:
        header = header_envelope[0].getchildren()[0]

    return body, header

# see http://www.w3.org/TR/2000/NOTE-SOAP-20000508/
# section 5.2.1 for an example of how the id and href attributes are used.
def resolve_hrefs(element, xmlids):
    for e in element:
        if e.get('id'):
            continue # don't need to resolve this element

        elif e.get('href'):
            resolved_element = xmlids[e.get('href').replace('#', '')]
            if resolved_element is None:
                continue
            resolve_hrefs(resolved_element, xmlids)

            # copies the attributes
            [e.set(k, v) for k, v in resolved_element.items()]

            # copies the children
            [e.append(child) for child in resolved_element.getchildren()]

            # copies the text
            e.text = resolved_element.text

        else:
            resolve_hrefs(e, xmlids)

    return element

def join_attachment(href_id, envelope, payload, prefix=True):
    '''
    Helper function for swa_to_soap.

    Places the data from an attachment back into a SOAP message, replacing
    its xop:Include element or href.

    @param  id        content-id or content-location of attachment
    @param  prefix    Set this to true if id is content-id or false if it is
                      content-location.  It prefixes a "cid:" to the href value.
    @param  envelope  soap envelope string to be operated on
    @param  payload   attachment data

    @return           tuple of length 2 with the new message and the
                      number of replacements made
    '''

    def replacing(parent, node, payload, numreplaces):
        if node.tag == '{%s}Include' % cocy.soaplib.ns_xop:
            attrib = node.attrib.get('href')
            if not attrib is None:
                if unquote(attrib) == href_id:
                    parent.remove(node)
                    parent.text = payload
                    numreplaces += 1
        else:
            for child in node:
                numreplaces = replacing(node, child, payload, numreplaces)

        return numreplaces

    # grab the XML element of the message in the SOAP body
    soaptree = etree.ElementTree.fromstring(envelope)
    soapbody = soaptree.find("{%s}Body" % cocy.soaplib.ns_soap_env)

    message = None
    for child in list(soapbody):
        if child.tag != "{%s}Fault" % cocy.soaplib.ns_soap_env:
            message = child
            break

    numreplaces = 0
    idprefix = ''

    if prefix == True:
        idprefix = "cid:"
    href_id = "%s%s" % (idprefix, href_id, )

    # Make replacement.
    for param in message:
        # Look for Include subelement.
        for sub in param:
            numreplaces = replacing(param, sub, payload, numreplaces)

        if numreplaces < 1:
            attrib = param.attrib.get('href')
            if not attrib is None:
                if unquote(attrib) == href_id:
                    del(param.attrib['href'])
                    param.text = payload
                    numreplaces += 1

    return (etree.ElementTree.tostring(soaptree), numreplaces)

def collapse_swa(content_type, envelope):
    '''
    Translates an SwA multipart/related message into an
    application/soap+xml message.

    References:
    SwA     http://www.w3.org/TR/SOAP-attachments
    XOP     http://www.w3.org/TR/xop10/
    MTOM    http://www.w3.org/TR/soap12-mtom/
            http://www.w3.org/Submission/soap11mtom10/

    @param  content_type value of the Content-Type header field, parsed by
                         cgi.parse_header() function
    @param  envelope     body of the HTTP message, a soap envelope
    @return              appication/soap+xml version of the given HTTP body
    '''

    # convert multipart messages back to pure SOAP
    mime_type = content_type[0]

    if 'multipart/related' not in mime_type:
        return envelope

    charset = content_type[1].get('charset', None)
    if charset is None:
        charset='ascii'

    # parse the body into an email.Message object
    msg_string = [
        "MIME-Version: 1.0",
        "Content-Type: %s; charset=%s" % (mime_type, charset),
        "",
        envelope
    ]

    msg = message_from_string('\r\n'.join(msg_string)) # our message

    soapmsg = None
    root = msg.get_param('start')

    # walk through sections, reconstructing pure SOAP
    for part in msg.walk():
        # skip the multipart container section
        if part.get_content_maintype() == 'multipart':
            continue

        # detect main soap section
        if (part.get('Content-ID') and part.get('Content-ID') == root) or \
           (root == None and part == msg.get_payload()[0]):
            soapmsg = part.get_payload()
            continue

        # binary packages
        cte = part.get("Content-Transfer-Encoding")

        payload = None
        if cte != 'base64':
            payload = b64encode(part.get_payload())
        else:
            payload = part.get_payload()

        cid = part.get("Content-ID").strip("<>")
        cloc = part.get("Content-Location")
        numreplaces = None

        # Check for Content-ID and make replacement
        if cid:
            soapmsg, numreplaces = join_attachment(cid, soapmsg, payload)

        # Check for Content-Location and make replacement
        if cloc and not cid and not numreplaces:
            soapmsg, numreplaces = join_attachment(cloc, soapmsg, payload, False)

    return soapmsg

def apply_mtom(headers, envelope, params, paramvals):
    '''
    Apply MTOM to a SOAP envelope, separating attachments into a
    MIME multipart message.

    References:
    XOP     http://www.w3.org/TR/xop10/
    MTOM    http://www.w3.org/TR/soap12-mtom/
            http://www.w3.org/Submission/soap11mtom10/

    @param headers   Headers dictionary of the SOAP message that would
                     originally be sent.
    @param envelope  SOAP envelope string that would have originally been sent.
    @param params    params attribute from the Message object used for the SOAP
    @param paramvals values of the params, passed to Message.to_xml
    @return          tuple of length 2 with dictionary of headers and
                     string of body that can be sent with HTTPConnection
    '''

    # grab the XML element of the message in the SOAP body
    soaptree = etree.ElementTree.fromstring(envelope)
    soapbody = soaptree.find("{%s}Body" % cocy.soaplib.ns_soap_env)

    message = None
    for child in list(soapbody):
        if child.tag != "%sFault" % (soaplib.ns_soap_env, ):
            message = child
            break

    # Get additional parameters from original Content-Type
    ctarray = []
    for n, v in headers.items():
        if n.lower() == 'content-type':
            ctarray = v.split(';')
            break
    roottype = ctarray[0].strip()
    rootparams = {}
    for ctparam in ctarray[1:]:
        n, v = ctparam.strip().split('=')
        rootparams[n] = v.strip("\"'")

    # Set up initial MIME parts.
    mtompkg = MIMEMultipart('related',
        boundary='?//<><>soaplib_MIME_boundary<>')
    rootpkg = None
    try:
        rootpkg = MIMEApplication(envelope, 'xop+xml', encode_7or8bit)
    except NameError:
        rootpkg = MIMENonMultipart("application", "xop+xml")
        rootpkg.set_payload(envelope)
        encode_7or8bit(rootpkg)

    # Set up multipart headers.
    del(mtompkg['mime-version'])
    mtompkg.set_param('start-info', roottype)
    mtompkg.set_param('start', '<soaplibEnvelope>')
    if 'SOAPAction' in headers:
        mtompkg.add_header('SOAPAction', headers.get('SOAPAction'))

    # Set up root SOAP part headers.
    del(rootpkg['mime-version'])

    rootpkg.add_header('Content-ID', '<soaplibEnvelope>')

    for n, v in rootparams.items():
        rootpkg.set_param(n, v)

    rootpkg.set_param('type', roottype)

    mtompkg.attach(rootpkg)

    # Extract attachments from SOAP envelope.
    for i in range(len(params)):
        name, typ = params[i]

        if typ == Attachment:
            id = "soaplibAttachment_%s" % (len(mtompkg.get_payload()), )

            param = message[i]
            param.text = ""

            incl = etree.ElementTree.SubElement(param, "{%s}Include" % cocy.soaplib.ns_xop)
            incl.attrib["href"] = "cid:%s" % id

            if paramvals[i].fileName and not paramvals[i].data:
                paramvals[i].load_from_file()

            data = paramvals[i].data
            attachment = None

            try:
                attachment = MIMEApplication(data, _encoder=encode_7or8bit)

            except NameError:
                attachment = MIMENonMultipart("application", "octet-stream")
                attachment.set_payload(data)
                encode_7or8bit(attachment)

            del(attachment['mime-version'])

            attachment.add_header('Content-ID', '<%s>' % (id, ))
            mtompkg.attach(attachment)

    # Update SOAP envelope.
    rootpkg.set_payload(etree.ElementTree.tostring(soaptree))

    # extract body string from MIMEMultipart message
    bound = '--%s' % (mtompkg.get_boundary(), )
    marray = mtompkg.as_string().split(bound)
    mtombody = bound
    mtombody += bound.join(marray[1:])

    # set Content-Length
    mtompkg.add_header("Content-Length", str(len(mtombody)))

    # extract dictionary of headers from MIMEMultipart message
    mtomheaders = {}
    for name, value in mtompkg.items():
        mtomheaders[name] = value

    if len(mtompkg.get_payload()) <= 1:
        return (headers, envelope)

    return (mtomheaders, mtombody)




#
# soaplib - Copyright (C) cocy.soaplib contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

import logging
logger = logging.getLogger(__name__)

import cgi
import shutil
import tempfile
import traceback

from lxml import etree

import cocy.soaplib

from cocy.soaplib.serializers.exception import Fault
from cocy.soaplib.serializers.primitive import string_encoding

from cocy.soaplib.soap import apply_mtom
from cocy.soaplib.soap import collapse_swa
from cocy.soaplib.soap import from_soap
from cocy.soaplib.util import reconstruct_url

HTTP_500 = '500 Internal server error'
HTTP_200 = '200 OK'
HTTP_405 = '405 Method Not Allowed'

class ValidationError(Fault):
    pass

def get_schema_node(pref, schema_nodes, types):
    """
    Return schema node for the given namespace prefix.
    types == None means the call is for creating a standalone xml schema file
                  for one single namespace.
    tyoes != None means the call is for creating the wsdl file.
    """

    # create schema node
    if not (pref in schema_nodes):
        if types is None:
            schema = etree.Element("{%s}schema" % cocy.soaplib.ns_xsd,
                                                    nsmap=soaplib.nsmap)
        else:
            schema = etree.SubElement(types, "{%s}schema" % cocy.soaplib.ns_xsd)

        schema.set("targetNamespace", cocy.soaplib.nsmap[pref])
        schema.set("elementFormDefault", "qualified")

        schema_nodes[pref] = schema

    else:
        schema = schema_nodes[pref]

    return schema

class Application(object):
    def __init__(self, services, tns, name=None, _with_partnerlink=False):
        '''
        @param A ServiceBase subclass that defines the exposed services.
        '''

        self.services = services
        self.__tns = tns
        self.__name = name
        self._with_plink = _with_partnerlink

        self.call_routes = {}
        self.__wsdl = None
        self.__public_methods = {}
        self.schema = self.build_schema()

    def get_name(self):
        """
        Returns service name that is seen in the name attribute of the
        definitions tag.
        """
        retval = self.__name

        if retval is None:
            retval = self.__class__.__name__.split('.')[-1]

        return retval

    name = property(get_name)

    def get_tns(self):
        """
        Returns default namespace that is seen in the targetNamespace attribute
        of the definitions tag.
        """
        retval = self.__tns

        if retval is None:
            service_name = self.get_name()

            if self.__class__.__module__ == '__main__':
                retval = '.'.join((service_name, service_name))
            else:
                retval = '.'.join((self.__class__.__module__, service_name))

            if retval.startswith('soaplib'):
                retval = self.services[0].get_tns()

        return retval

    tns = property(get_tns)

    def __build_schema_nodes(self, schema_entries, types=None):
        """
        Fill individual <schema> nodes for every service that are part of this
        app.
        """

        schema_nodes = {}

        for pref in schema_entries.namespaces:
            schema = get_schema_node(pref, schema_nodes, types)

            # append import tags
            for namespace in schema_entries.imports[pref]:
                import_ = etree.SubElement(schema, "{%s}import"% cocy.soaplib.ns_xsd)
                import_.set("namespace", namespace)
                if types is None:
                    import_.set('schemaLocation', "%s.xsd" %
                                        cocy.soaplib.get_namespace_prefix(namespace))

            # append element tags
            for node in schema_entries.namespaces[pref].elements.values():
                schema.append(node)

            # append simpleType and complexType tags
            for node in schema_entries.namespaces[pref].types.values():
                schema.append(node)

        return schema_nodes

    def build_schema(self, types=None):
        """
        Unify the <schema> nodes required for this app.
        """

        if types is None:
            # populate call routes
            for s in self.services:
                s.__tns__ = self.get_tns()
                inst = self.get_service(s)

                for method in inst.public_methods:
                    method_name = "{%s}%s" % (self.get_tns(), method.name)

                    if method_name in self.call_routes:
                        o = self.call_routes[method_name]
                        raise Exception("%s.%s.%s overwrites %s.%s.%s" %
                                        (s.__module__, s.__name__, method.name,
                                         o.__module__, o.__name__, method.name))

                    else:
                        logger.debug('adding method %r' % method_name)
                        self.call_routes[method_name] = s
                        self.call_routes[method.name] = s

        # populate types
        schema_entries = None
        for s in self.services:
            inst = self.get_service(s)
            schema_entries = inst.add_schema(schema_entries)

        schema_nodes = self.__build_schema_nodes(schema_entries, types)

        return schema_nodes

    def get_service_class(self, method_name):
        """
        This call maps method names to the services that will handle them.
        Override this function to alter the method mappings. Just try not to get
        too crazy with regular expressions :)
        """
        return self.call_routes[method_name]

    def get_service(self, service, http_req_env=None):
        """
        The function that maps service classes to service instances. Overriding
        this function is useful in case e.g. you need to pass additional
        parameters to service constructors.
        """
        return service(http_req_env)

    def get_schema(self):
        """
        Simple accessor method that caches application's xml schema, once
        generated.
        """
        if self.schema is None:
            return self.build_schema()
        else:
            return self.schema

    def get_wsdl(self, url):
        """
        Simple accessor method that caches the wsdl of the application, once
        generated.
        """
        if self.__wsdl is None:
            return self.__build_wsdl(url)
        else:
            return self.__wsdl

    def __is_wsdl_request(self, req_env):
        # Get the wsdl for the service. Assume path_info matches pattern:
        # /stuff/stuff/stuff/serviceName.wsdl or
        # /stuff/stuff/stuff/serviceName/?wsdl

        return (
                req_env['REQUEST_METHOD'].lower() == 'get'
            and (
                   req_env['QUERY_STRING'].endswith('wsdl')
                or req_env['PATH_INFO'].endswith('wsdl')
            )
        )

    def __build_wsdl(self, url):
        """
        Build the wsdl for the application.
        """
        ns_wsdl = cocy.soaplib.ns_wsdl
        ns_soap = cocy.soaplib.ns_soap
        ns_plink = cocy.soaplib.ns_plink

        ns_tns = self.get_tns()
        pref_tns = cocy.soaplib.get_namespace_prefix(ns_tns) #'tns'
        # FIXME: this can be enabled when cocy.soaplib.nsmap is no longer global
        #soaplib.set_namespace_prefix(ns_tns, pref_tns)

        # FIXME: doesn't look so robust
        url = url.replace('.wsdl', '')

        service_name = self.get_name()

        # create wsdl root node
        root = etree.Element("{%s}definitions" % ns_wsdl, nsmap=soaplib.nsmap)
        root.set('targetNamespace', ns_tns)
        root.set('name', service_name)

        # create types node
        types = etree.SubElement(root, "{%s}types" % ns_wsdl)

        self.build_schema(types)
        messages = set()

        for s in self.services:
            s=self.get_service(s,None)

            s.add_messages_for_methods(root, messages)

        if self._with_plink:
            # create plink node
            plink = etree.SubElement(root, '{%s}partnerLinkType' % ns_plink)
            plink.set('name', service_name)
            self.add_partner_link(root, service_name, types, url, plink)

        # create service node
        service = etree.SubElement(root, '{%s}service' % ns_wsdl)
        service.set('name', service_name)
        self.add_service(root, service_name, types, url, service)

        # create portType node
        port_type = etree.SubElement(root, '{%s}portType' % ns_wsdl)
        port_type.set('name', service_name)

        # create binding nodes
        binding = etree.SubElement(root, '{%s}binding' % ns_wsdl)
        binding.set('name', service_name)
        binding.set('type', '%s:%s'% (pref_tns, service_name))

        soap_binding = etree.SubElement(binding, '{%s}binding' % ns_soap)
        soap_binding.set('style', 'document')
        soap_binding.set('transport', 'http://schemas.xmlsoap.org/soap/http')

        for s in self.services:
            s = self.get_service(s)
            s.add_port_type(root, service_name, types, url, port_type)
            s.add_bindings_for_methods(root, service_name, types, url, binding)

        self.__wsdl = etree.tostring(root, xml_declaration=True,
                                                               encoding="UTF-8")

        return self.__wsdl

    def add_partner_link(self, root, service_name, types, url, plink):
        """
        Add the partnerLinkType node to the wsdl.
        """
        ns_plink = cocy.soaplib.ns_plink
        pref_tns = cocy.soaplib.get_namespace_prefix(self.get_tns())

        role = etree.SubElement(plink, '{%s}role' % ns_plink)
        role.set('name', service_name)

        plink_port_type = etree.SubElement(role, '{%s}portType' % ns_plink)
        plink_port_type.set('name', '%s:%s' % (pref_tns, service_name))

    def add_service(self, root, service_name, types, url, service):
        """
        Add service node to the wsdl.
        """
        pref_tns = cocy.soaplib.get_namespace_prefix(self.get_tns())

        wsdl_port = etree.SubElement(service, '{%s}port' % cocy.soaplib.ns_wsdl)
        wsdl_port.set('name', service_name)
        wsdl_port.set('binding', '%s:%s' % (pref_tns, service_name))

        addr = etree.SubElement(wsdl_port, '{%s}address' % cocy.soaplib.ns_soap)
        addr.set('location', url)

    def __handle_wsdl_request(self, req_env, start_response, url):
        http_resp_headers = {'Content-Type': 'text/xml'}

        try:
            self.get_wsdl(url)
            self.on_wsdl(req_env, self.__wsdl) # implementation hook

            http_resp_headers['Content-Length'] = str(len(self.__wsdl))
            start_response(HTTP_200, http_resp_headers.items())

            return [self.__wsdl]

        except Exception, e:
            logger.error(traceback.format_exc())

            # implementation hook
            self.on_wsdl_exception(req_env, e)

            start_response(HTTP_500, http_resp_headers.items())

            return [""]

    def __decode_soap_request(self, http_env, http_payload):
        """
        Decode http payload using information in the http header
        """

        # fyi, here's what the parse_header function returns:
        # >>> import cgi; cgi.parse_header("text/xml; charset=utf-8")
        # ('text/xml', {'charset': 'utf-8'})
        content_type = cgi.parse_header(http_env.get("CONTENT_TYPE"))
        charset = content_type[1].get('charset',None)
        if charset is None:
            charset = 'ascii'

        http_payload = collapse_swa(content_type, http_payload)

        # deserialize the body of the message
        req_payload, req_header = from_soap(http_payload, charset)

        return req_header, req_payload

    def validate_request(self, payload):
        """
        Method to be overriden to perform any sort of custom input validation.
        """
        pass

    def __get_method_name(self, http_req_env, soap_req_payload):
        """
        Guess method name basing on various information in the request.
        """
        retval = None

        if soap_req_payload is not None:
            retval = soap_req_payload.tag
            logger.debug("\033[92mMethod name from xml tag: %r\033[0m" % retval)
        else:
            # check HTTP_SOAPACTION
            retval = http_req_env.get("HTTP_SOAPACTION")

            if retval is not None:
                if retval.startswith('"') and retval.endswith('"'):
                    retval = retval[1:-1]

                if retval.find('/') >0:
                    retvals = retval.split('/')
                    retval = '{%s}%s' % (retvals[0], retvals[1])

                logger.debug("\033[92m"
                             "Method name from HTTP_SOAPACTION: %r"
                             "\033[0m" % retval)

        return retval

    def __handle_soap_request(self, req_env, start_response, url):
        """
        This function is too big.
        """

        http_resp_headers = {
            'Content-Type': 'text/xml',
            'Content-Length': '0',
        }
        method_name = None

        try:
            # implementation hook
            self.on_call(req_env)

            if req_env['REQUEST_METHOD'].lower() != 'post':
                http_resp_headers['Allow'] = 'POST'
                start_response(HTTP_405, http_resp_headers.items())
                return ['']

            input = req_env.get('wsgi.input')
            length = req_env.get("CONTENT_LENGTH")
            body = input.read(int(length))

            try:
                service = None
                soap_req_header, soap_req_payload = self.__decode_soap_request(
                                                                req_env, body)
                if not (soap_req_payload is None):
                    self.validate_request(soap_req_payload)

                method_name = self.__get_method_name(req_env, soap_req_payload)
                if method_name is None:
                    resp = "Could not extract method name from the request!"
                    http_resp_headers['Content-Length'] = str(len(resp))
                    start_response(HTTP_500, http_resp_headers.items())
                    return [resp]

                service_class = self.get_service_class(method_name)
                service = self.get_service(service_class, req_env)

            finally:
                # for performance reasons, we don't want the following to run
                # in production even though we won't see the results.
                if logger.level == logging.DEBUG:
                    try:
                        logger.debug(etree.tostring(etree.fromstring(body),
                                                             pretty_print=True))
                    except etree.XMLSyntaxError,e:
                        logger.debug(body)
                        raise Fault('Client.XMLSyntax', 'Error at line: %d, col: %d'
                                                                    % e.position)

            # retrieve the method descriptor
            descriptor = service.get_method(method_name)
            func = getattr(service, descriptor.name)

            # decode header object
            if soap_req_header is not None and len(soap_req_header) > 0:
                in_header = descriptor.in_header
                service.soap_in_header = in_header.from_xml(soap_req_header)

            # decode method arguments
            if soap_req_payload is not None and len(soap_req_payload) > 0:
                params = descriptor.in_message.from_xml(soap_req_payload)
            else:
                params = [None] * len(descriptor.in_message._type_info)

            # implementation hook
            service.on_method_call(req_env, method_name, params,
                                                               soap_req_payload)

            # call the method
            result_raw = service.call_wrapper(func, params)

            # construct the soap response, and serialize it
            envelope = etree.Element('{%s}Envelope' % cocy.soaplib.ns_soap_env,
                                                            nsmap=soaplib.nsmap)

            #
            # header
            #
            soap_header_elt = etree.SubElement(envelope,
                                             '{%s}Header' % cocy.soaplib.ns_soap_env)

            if service.soap_out_header != None:
                if descriptor.out_header is None:
                    logger.warning("Skipping soap response header as %r method "
                                   "is not published to have a soap response "
                                   "header" % method_name)
                else:
                    descriptor.out_header.to_xml(
                        service.soap_out_header,
                        self.get_tns(),
                        soap_header_elt,
                        descriptor.out_header.get_type_name()
                    )

            if len(soap_header_elt) > 0:
                envelope.append(soap_header_elt)

            #
            # body
            #
            soap_body = etree.SubElement(envelope,
                                               '{%s}Body' % cocy.soaplib.ns_soap_env)

            # instantiate the result message
            result_message = descriptor.out_message()

            # assign raw result to its wrapper, result_message
            out_type = descriptor.out_message._type_info

            if len(out_type) > 0:
                if len(out_type) == 1:
                    attr_name = descriptor.out_message._type_info.keys()[0]
                    setattr(result_message, attr_name, result_raw)
                else:
                    for i in range(len(out_type)):
                        attr_name = descriptor.out_message._type_info.keys()[i]
                        setattr(result_message, attr_name, result_raw[i])

            # transform the results into an element
            descriptor.out_message.to_xml(result_message, self.get_tns(),
                                                                      soap_body)

            # implementation hook
            service.on_method_return(req_env, result_raw, soap_body,
                                                              http_resp_headers)

            #
            # misc
            #
            results_str = etree.tostring(envelope, xml_declaration=True,
                                                       encoding=string_encoding)

            if descriptor.mtom:
                http_resp_headers, results_str = apply_mtom(http_resp_headers,
                    results_str, descriptor.out_message._type_info,[result_raw])

            # implementation hook
            self.on_return(req_env, http_resp_headers, results_str)

            # initiate the response
            http_resp_headers['Content-Length'] = str(len(results_str))
            start_response(HTTP_200, http_resp_headers.items())

            if logger.level == logging.DEBUG:
                logger.debug('\033[91m'+ "Response" + '\033[0m')
                logger.debug(etree.tostring(envelope, xml_declaration=True,
                                                             pretty_print=True))

            # return the serialized results
            return [results_str]

        # The user issued a Fault, so handle it just like an exception!
        except Fault, e:
            return self.__handle_fault(req_env, start_response,
                                                http_resp_headers, service, e)

        except Exception, e:
            fault = Fault('Server', str(e))

            return self.__handle_fault(req_env, start_response,
                                              http_resp_headers, service, fault)

    def __handle_fault(self, req_env, start_response, http_resp_headers,
                                                                  service, exc):
        stacktrace=traceback.format_exc()
        logger.error(stacktrace)

        # implementation hook
        if not (service is None):
            service.on_method_exception_object(req_env, exc)
        self.on_exception_object(req_env, exc)

        # FIXME: There's no way to alter soap response headers for the user.
        envelope = etree.Element('{%s}Envelope' % cocy.soaplib.ns_soap_env)
        body = etree.SubElement(envelope, '{%s}Body' % cocy.soaplib.ns_soap_env,
                                                            nsmap=soaplib.nsmap)
        exc.__class__.to_xml(exc, self.get_tns(), body)

        if not (service is None):
            service.on_method_exception_xml(req_env, body)
        self.on_exception_xml(req_env, body)

        if logger.level == logging.DEBUG:
            logger.debug(etree.tostring(envelope, pretty_print=True))

        # initiate the response
        fault_str = etree.tostring(envelope)
        http_resp_headers['Content-Length'] = str(len(fault_str))
        start_response(HTTP_500, http_resp_headers.items())

        return [fault_str]

    def __call__(self, req_env, start_response, wsgi_url=None):
        '''
        This method conforms to the WSGI spec for callable wsgi applications
        (PEP 333). It looks in environ['wsgi.input'] for a fully formed soap
        request envelope, will deserialize the request parameters and call the
        method on the object returned by the get_handler() method.

        @param the http environment
        @param a callable that begins the response message
        @param the optional url
        @returns the string representation of the soap call
        '''

        url = wsgi_url
        if url is None:
            url = reconstruct_url(req_env).split('.wsdl')[0]

        if self.__is_wsdl_request(req_env):
            return self.__handle_wsdl_request(req_env, start_response, url)
        else:
            return self.__handle_soap_request(req_env, start_response, url)

    def on_call(self, environ):
        '''
        This is the first method called when this WSGI app is invoked
        @param the wsgi environment
        '''
        pass

    def on_wsdl(self, environ, wsdl):
        '''
        This is called when a wsdl is requested

        @param the wsgi environment
        @param the wsdl string
        '''
        pass

    def on_wsdl_exception(self, environ, exc):
        '''
        Called when an exception occurs durring wsdl generation

        @param the wsgi environment
        @param exc the exception
        @param the fault response string
        '''
        pass

    def on_return(self, environ, http_headers, return_str):
        '''
        Called before the application returns

        @param the wsgi environment
        @param http response headers as dict
        @param return string of the soap request
        '''
        pass

    def on_exception_object(self, environ, exc):
        '''
        Called when the app throws an exception. (might be inside or outside the
        service call.
        @param the wsgi environment
        @param the fault object
        '''
        pass

    def on_exception_xml(self, environ, fault_xml):
        '''
        Called when the app throws an exception. (might be inside or outside the
        service call.
        @param the wsgi environment
        @param the xml element containing the xml serialization of the fault
        '''
        pass

class ValidatingApplication(Application):
    def build_schema(self, types=None):
        """
        Build application schema specifically for xml validation purposes.
        """
        schema_nodes = Application.build_schema(self, types)

        if types is None:
            pref_tns = cocy.soaplib.get_namespace_prefix(self.get_tns())
            logger.debug("generating schema for targetNamespace=%r, prefix: %r"
                                                   % (self.get_tns(), pref_tns))

            tmp_dir_name = tempfile.mkdtemp()

            # serialize nodes to files
            for k,v in schema_nodes.items():
                file_name = '%s/%s.xsd' % (tmp_dir_name, k)
                f = open(file_name, 'w')
                etree.ElementTree(v).write(f, pretty_print=True)
                f.close()
                logger.debug("writing %r for ns %s" % (file_name,
                                                            cocy.soaplib.nsmap[k]))

            f = open('%s/%s.xsd' % (tmp_dir_name, pref_tns), 'r')

            logger.debug("building schema...")
            self.schema = etree.XMLSchema(etree.parse(f))

            logger.debug("schema %r built, cleaning up..." % self.schema)
            f.close()
            shutil.rmtree(tmp_dir_name)
            logger.debug("removed %r" % tmp_dir_name)

        return self.schema

    def validate_request(self, payload):
        schema = self.schema
        ret = schema.validate(payload)

        logger.debug("validation result: %s" % str(ret))
        if ret == False:
            err = schema.error_log.last_error

            fault_code = 'Client.SchemaValidation'

            raise ValidationError(fault_code, faultstring=str(err))




#
# soaplib - Copyright (C) cocy.soaplib contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

# namespace map

ns_xsd = 'http://www.w3.org/2001/XMLSchema'
ns_xsi = 'http://www.w3.org/2001/XMLSchema-instance'
ns_plink = 'http://schemas.xmlsoap.org/ws/2003/05/partner-link/'
ns_soap = 'http://schemas.xmlsoap.org/wsdl/soap/'
ns_wsdl = 'http://schemas.xmlsoap.org/wsdl/'
ns_soap_enc = 'http://schemas.xmlsoap.org/soap/encoding/'
ns_soap_env = 'http://schemas.xmlsoap.org/soap/envelope/'
ns_soap12_env = 'http://www.w3.org/2003/05/soap-envelope/'
ns_soap12_enc = 'http://www.w3.org/2003/05/soap-encoding/'
ns_wsa = 'http://schemas.xmlsoap.org/ws/2003/03/addressing'
ns_xop = 'http://www.w3.org/2004/08/xop/include'

nsmap = {
    'xs': ns_xsd,
    'xsi': ns_xsi,
    'plink': ns_plink,
    'soap': ns_soap,
    'wsdl': ns_wsdl,
    'senc': ns_soap_enc,
    'senv': ns_soap_env,
    's12env': ns_soap12_env,
    's12enc': ns_soap12_enc,
    'wsa': ns_wsa,
    'xop': ns_xop,
}

# prefix map
prefmap = dict([(b,a) for a,b in nsmap.items()])

const_prefmap = dict(prefmap)
const_nsmap = dict(nsmap)

_ns_counter = 0
def get_namespace_prefix(ns):
    global _ns_counter

    assert ns != "__main__"
    assert ns != "soaplib.serializers.base"

    assert (isinstance(ns, str) or isinstance(ns, unicode)), ns

    if not (ns in prefmap):
        pref = "s%d" % _ns_counter
        while pref in nsmap:
            _ns_counter += 1
            pref = "s%d" % _ns_counter

        prefmap[ns] = pref
        nsmap[pref] = ns

        _ns_counter += 1

    else:
        pref = prefmap[ns]

    return pref

def set_namespace_prefix(ns, pref):
    if pref in nsmap and nsmap[pref] != ns:
        ns_old = nsmap[pref]
        del prefmap[ns_old]
        get_namespace_prefix(ns_old)

    cpref = get_namespace_prefix(ns)
    del nsmap[cpref]

    prefmap[ns] = pref
    nsmap[pref] = ns
